"""
Moments for One-Factor SV
"""
import math
from fractions import Fraction as Frac

from ajdmom.poly import Poly
from ajdmom.ito_mom import moment_v, moment_IEII
from ajdmom.utils import comb
from ajdmom.mdl_1fsv.cmom import ve_IEII


def b_n(n1, n2, n3, n4, n5):
    r"""Represent :math:`b_2(\boldsymbol{n})` in :eq:`b2-n` in Poly

    :param int n1: times of constant being selected.
    :param int n2: times of :math:`v_{n-1}` being selected.
    :param int n3: times of :math:`eI_n` being selected.
    :param int n4: times of :math:`I_n` being selected.
    :param int n5: times of :math:`I_n^{*}` being selected.
    :return: a poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)').
    :rtype: Poly
    """
    b = Poly()
    kf = ['e^{-kh}', 'h', 'k^{-}', 'mu', 'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)']
    b.set_keyfor(kf)
    #
    for i1 in range(n1, -1, -1):  # i1: mu * h
        for i2 in range(n1 - i1, -1, -1):  # i2: -theta/2 * h
            for i3 in range(n1 - i1 - i2, -1, -1):  # i3: 1/(2k) * theta
                i4 = n1 - i1 - i2 - i3  # i4: -e^{-kh}/(2k) * theta
                for i in range(n2 + 1):  # i : -e^{-kh}/(2k)
                    for j in range(n4 + 1):  # j : -sigma_v/(2k)
                        num = comb(n1, [i1, i2, i3, i4])
                        num *= math.comb(n2, i)
                        num *= math.comb(n4, j)
                        #   e^{-kh},h    ,k^{-}        ,mu,theta,sigma_v,rho,sqrt(1-rho^2)
                        key = (i4 + i, i1 + i2, i3 + i4 + n2 + n3 + j, i1, n1 - i1, n3 + j, n4 - j, n5)
                        num *= (-1) ** (i2 + i4 + n2 + i + j)
                        den = 2 ** (n1 - i1 + n2 + n3 + j)
                        b.add_keyval(key, Frac(num, den))
    return b


def moment_comb(n, n1, n2, n3, n4, n5):
    r"""Moment for this combination in expansion of :math:`y_n^l`

    :param int n: l in :math:`E[\overline{y}_{n}^l]`.
    :param int n1: times of constant being selected.
    :param int n2: times of :math:`v_{n-1}` being selected.
    :param int n3: times of :math:`eI_n` being selected.
    :param int n4: times of :math:`I_n` being selected.
    :param int n5: times of :math:`I_n^{*}` being selected.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','v_{n-1}','k^{-}','mu','theta','sigma_v','rho',
       'sqrt(1-rho^2)').
    :rtype: Poly
    """
    poly = ve_IEII(n2, n3, n4, n5)
    # ['e^{-kh}','h','v_{n-1}','k^{-}','theta','sigma_v']
    b = b_n(n1, n2, n3, n4, n5)
    # ['e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)']
    #
    keyfor = ['e^{-kh}', 'h', 'v_{n-1}', 'k^{-}', 'mu', 'theta', 'sigma_v', 'rho',
              'sqrt(1-rho^2)']
    #
    keyIndexes = ([0, 1, 2, 3, -1, 4, 5, -1, -1], [0, 1, -1, 2, 3, 4, 5, 6, 7])
    # -1 for not having this term
    poly = poly.mul_poly(b, keyIndexes, keyfor)
    #
    c = comb(n, [n1, n2, n3, n4, n5])
    return c * poly


def sub_v(poly):
    """Substitute :math:`v_{n-1}` with its moment (poly) in Moment of y_n

    :param Poly poly: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','v_{n-1}','k^{-}','mu','theta','sigma_v','rho',
       'sqrt(1-rho^2)').
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)').
    :rtype: Poly
    """
    poly_sum = Poly()
    kf = ['e^{-kh}', 'h', 'k^{-}', 'mu', 'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)']
    poly_sum.set_keyfor(kf)
    #
    for k in poly:
        poln = moment_v(k[2])  # ['theta','sigma_v^2/k']
        for key in poln:
            i, j = key
            knw = (k[0], k[1], k[3] + j, k[4], k[5] + i, k[6] + 2 * j, k[7], k[8])
            val = poln[key] * poly[k]
            poly_sum.add_keyval(knw, val)
    return poly_sum


def moment_y(l):
    """Moment of :math:`y_n` of order :math:`l`

    :param int l: order of the moment.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)').
    :rtype: Poly
    """
    poly = Poly()
    kf = ['e^{-kh}', 'h', 'v_{n-1}', 'k^{-}', 'mu', 'theta', 'sigma_v', 'rho',
          'sqrt(1-rho^2)']
    poly.set_keyfor(kf)
    for n1 in range(l, -1, -1):
        for n2 in range(l - n1, -1, -1):
            for n3 in range(l - n1 - n2, -1, -1):
                for n4 in range(l - n1 - n2 - n3, -1, -1):
                    n5 = l - n1 - n2 - n3 - n4
                    poly.merge(moment_comb(l, n1, n2, n3, n4, n5))
    poly_sum = sub_v(poly)
    poly_sum.remove_zero()
    return poly_sum


##########
# scalar and (partial) derivative
##########
def dpoly(poly, wrt):
    """Partial derivative of central moment w.r.t. parameter wrt

    :param Poly poly: central moment represented by the poly whose attribute
       ``keyfor`` =
       ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)').
    :param str wrt: with respect to.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)').
    :rtype: Poly
    """
    pold = Poly()
    kf = ('e^{-kh}', 'h', 'k^{-}', 'mu', 'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)')
    pold.set_keyfor(kf)
    #
    # partial derivative w.r.t. k
    if wrt == 'k':
        for k in poly:
            if k[0] != 0:
                knw = list(k)
                knw[1] += 1
                val = (-k[0]) * poly[k]
                pold.add_keyval(tuple(knw), val)
            if k[2] != 0:
                knw = list(k)
                knw[2] += 1
                val = (-k[2]) * poly[k]
                pold.add_keyval(tuple(knw), val)
    elif wrt in ['mu', 'theta', 'sigma_v']:
        if wrt == 'mu': i = 3
        if wrt == 'theta': i = 4
        if wrt == 'sigma_v': i = 5
        for k in poly:
            if k[i] != 0:
                knw = list(k)
                knw[i] -= 1
                val = k[i] * poly[k]
                pold.add_keyval(tuple(knw), val)
    # partial derivative w.r.t. rho
    elif wrt == 'rho':
        for k in poly:
            if k[6] != 0:
                knw = list(k)
                knw[6] -= 1
                val = k[6] * poly[k]
                pold.add_keyval(tuple(knw), val)
            if k[7] != 0:
                knw = list(k)
                knw[6] += 1
                knw[7] -= 2
                val = (-k[7]) * poly[k]
                pold.add_keyval(tuple(knw), val)
    else:
        msg = "wrt must be either 'k','mu','theta','sigma_v' or 'rho'!"
        raise ValueError(msg)
    return pold


def poly2num(poly, par):
    """Decode poly back to scalar

    :param Poly poly: poly to be decoded with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)').
    :param dict par: parameters in dict.
    :return: scalar of the poly.
    :rtype: float
    """
    k = par['k']
    h = par['h']
    mu = par['mu']
    theta = par['theta']
    sigma_v = par['sigma_v']
    rho = par['rho']
    #
    value = 0
    for K in poly:
        val = poly[K] * math.exp(-K[0] * k * h) * (h ** K[1]) * (k ** (-K[2]))
        val *= (mu ** K[3])
        val *= (theta ** K[4]) * (sigma_v ** K[5]) * (rho ** K[6])
        val *= (1 - rho ** 2) ** (K[7] / 2)
        value += val
    return value


def m(l, par):
    """Moment in scalar

    :param int l: order of the moment.
    :param dict par: parameters in dict.
    :return: scalar of the moment.
    :rtype: float
    """
    moment = moment_y(l)
    value = poly2num(moment, par)
    return value


def dm(l, par, wrt):
    """Partial derivative of moment w.r.t. parameter wrt

    :param int l: order of the moment.
    :param dict par: parameters in dict.
    :param str wrt: with respect to.
    :return: scalar of the partial derivative.
    :rtype: float
    """
    moment = moment_y(l)
    pold = dpoly(moment, wrt)
    value = poly2num(pold, par)
    return value


if __name__ == "__main__":
    # Example usage of the module, see 'tests/test_mdl_1fsv.py' for more test
    from pprint import pprint
    import sys

    print('\nExample usage of the module function\n')
    args = sys.argv[1:]
    n = 2 if len(args) == 0 else int(args[0])
    mom = moment_y(n)
    print(f"moment_y({n}) = \n")
    pprint(mom)
    print(f"\nwhich is a Poly with attribute keyfor = \n{mom.keyfor}")
    #
    par = {'mu': 0.125, 'k': 0.1, 'theta': 0.25, 'sigma_v': 0.1, 'rho': -0.7, 'h': 1}
    val = poly2num(mom, par)
    print(f"\nnumerical value = {val}, given parameters = \n{par}")


import numpy as np
import math

from ajdmom.mdl_1fsvj.euler import r1FSVJ
from ajdmom.mdl_1fsvj.mom import m
from ajdmom.mdl_1fsvj.cmom import cm
from ajdmom.mdl_1fsvj.cov import cov


def comp_moments(orders, par, y_series):
    mnames = []
    tms = []
    sms = []
    for n in orders:
        tm = m(n, par)  # theoretial moment calculated by the package
        sm = np.mean(y_series ** n)  # sample moment
        mnames.append(f'moment({n})')
        tms.append(tm)
        sms.append(sm)
    return ((mnames, tms, sms))


def comp_cmoments(orders, par, y_series):
    mnames = []
    tms = []
    sms = []
    diff = y_series - np.mean(y_series)
    for n in orders:
        tm = cm(n, par)  # theoretial cmoment calculated by the package
        sm = np.mean(diff ** n)  # sample cmoment
        mnames.append(f'cmoment({n})')
        tms.append(tm)
        sms.append(sm)
    return ((mnames, tms, sms))


def comp_cov(order_pairs, par, y_series):
    mnames = []
    tms = []
    sms = []
    N = len(y_series)
    pre = y_series[0:(N - 1)]
    nxt = y_series[1:N]
    for nm in order_pairs:
        n, m = nm
        tm = cov(n, m, par)  # theoretial cov calculated by the package
        sm = np.cov(pre ** n, nxt ** m)[0][1]  # sample cov
        mnames.append(f'cov({n},{m})')
        tms.append(tm)
        sms.append(sm)
    return ((mnames, tms, sms))


def print_comp(comp):
    txt = ''.ljust(10)
    txt += 'derived moment'.rjust(15) + 'sample moment'.rjust(15)
    txt += 'diff'.rjust(12) + 'diff(in %)'.rjust(12) + '\n'
    #
    for i in range(len(comp[0])):
        name = comp[0][i]
        tm = comp[1][i]
        sm = comp[2][i]
        diff = abs(tm - sm)
        if tm == 0:
            diff_percent = math.nan
        else:
            diff_percent = diff / abs(tm)
        #
        txt += '{:<10}{:>15.4f}{:>15.4f}{:>12.4f}{:>12.0%}\n'.format(name,
                                                                     tm, sm, diff, diff_percent)
    print(txt)


# h should be in par, since it is needed for m(), cm(), cov()
par = {'mu': 0.125, 'k': 0.1, 'theta': 0.25, 'sigma_v': 0.1, 'rho': -0.7,
       'lambda': 0.01, 'mu_j': 0, 'sigma_j': 0.05, 'h': 1}
v_0 = par['theta']
y_series = r1FSVJ(v_0, par, N=4000 * 1000, n_segment=10, h=par['h'])

orders = [1, 2, 3, 4, 5]
mom_comp = comp_moments(orders, par, y_series)
print_comp(mom_comp)

orders = [1, 2, 3, 4, 5]
cmom_comp = comp_cmoments(orders, par, y_series)
print_comp(cmom_comp)

orders = [(1, 1), (2, 1), (1, 2), (3, 1), (2, 2), (1, 3), (4, 1), (3, 2), (2, 3), (1, 4)]
cov_comp = comp_cov(orders, par, y_series)
print_comp(cov_comp)



r"""
Itô process conditional moments under Square-Root Jump-Diffusion Process

Condition on the starting variance and jumps occurred over the interval.

Note that

- The unconditional moment derivation is supported within
  :py:mod:`ajdmom.mdl_srjd.mom` and :py:mod:`ajdmom.mdl_srjd.cmom`.

- The conditional (given :math:`v_0`) moment derivation is supported within
  :py:mod:`ajdmom.mdl_srjd.cond_mom` and :py:mod:`ajdmom.mdl_srjd.cond_cmom`.


Highlights
===========

- Offer supports for deriving the conditional moments for models including
  jumps in the variance (
  :abbr:`SRJD(Square-Root Jump Diffusion)`,
  :abbr:`SVVJ(Stochastic Volatility with Jumps in the Variance)`,
  :abbr:`SVIJ(Stochastic Volatility with Independent Jumps in the price and
  variance)` and
  :abbr:`SVCJ(Stochastic Volatility with Contemporaneous Jumps in the price
  and variance)`).

- The conditions are that current variance and jumps over the interval are
  given beforehand.


Square-Root Jump Diffusion
============================

The Square-Root Jump Diffusion (SRJD) is described by the following
:abbr:`SDE(Stochastic Differential Equation)`,

.. math::

   dv(t) = k(\theta - v(t))dt + \sigma_v\sqrt{v(t)}dw^v(t) + dz(t),

which adds a jump component :math:`z(t)`
(a :abbr:`CPP(Compound Poisson Process)`) into the CIR diffusion.
We introduce the following notations for simplification,

.. math::

   \begin{align*}
   I\!Z_t &\triangleq \int_0^tdz(s)
   ~~\left(\equiv \sum_{i=1}^{N(t)} J_i\right),\\
   I\!E\!Z_t &\triangleq \int_0^te^{ks}dz(s)
   ~~\left(\equiv \sum_{i=1}^{N(t)} e^{ks_i}J_i\right).
   \end{align*}

The solution to the SDE can then be expressed as

.. math::

  e^{kt}v(t)  = (v_0-\theta) + e^{kt}\theta + \sigma_v I\!E_t + I\!E\!Z_t,

noting that :math:`v_0 \equiv v(0)` and
:math:`I\!E_t \equiv \int_0^t e^{ks}\sqrt{v(s)} dw^v(s)`.
Further,

.. math::

   e^{kt}(v(t) - \theta) - (v_0-\theta) = \sigma_v I\!E_t + I\!E\!Z_t.

In order to derive moment formulas for models including jumps in
the variance, SVVJ, SVIJ and SVCJ, we first compute the condtional
moments

.. math::

   \mathbb{E}[I\!E_t^{n_1} I_t^{n_2} (I_t^{*})^{n_3}|v_0, z(s), 0\le s\le t],

noting that :math:`I_t \equiv \int_0^t \sqrt{v(s)} dw^v(s)`,
:math:`I_t^{*} \equiv \int_0^t \sqrt{v(s)} dw(s)` and the Brownian motion in
the price process is decomposed as
:math:`w^s(t) = \rho w^v(t) + \sqrt{1-\rho^2}w(t)`, refer to the
:doc:`../theory` page
for the definitions of these quantities.

Recursive Equations
====================

Itô process moment
------------------------

.. math::
  :label: ito-jmp-moment-ii

  \begin{align*}
  &\mathbb{E}[ I\!E_t^{n_1} I_t^{n_2} (I_t^{*})^{n_3}|v_0, z(s), 0\le s\le t ] \\
  &= \frac{1}{2} n_1(n_1-1)(v_0-\theta)\times&\int_0^t e^{ks} \mathbb{E}[ I\!E_s^{n_1-2}I_s^{n_2}(I_s^{*})^{n_3}]ds\\
  &\quad + \frac{1}{2} n_1(n_1-1)\theta    \times&\int_0^t e^{2ks} \mathbb{E}[ I\!E_s^{n_1-2}I_s^{n_2}(I_s^{*})^{n_3}]ds\\
  &\quad + \frac{1}{2} n_1(n_1-1)\sigma_v   \times&\int_0^t e^{ks} \mathbb{E}[ I\!E_s^{n_1-1}I_s^{n_2}(I_s^{*})^{n_3}]ds\\
  &\quad + \frac{1}{2} n_1(n_1-1)       \times&\int_0^t e^{ks}I\!E\!Z_s \mathbb{E}[ I\!E_s^{n_1-2}I_s^{n_2}(I_s^{*})^{n_3}]ds\\
  &\color{blue}\quad + \frac{1}{2} n_2(n_2-1)(v_0-\theta)  \times&\color{blue}\int_0^t e^{-ks} \mathbb{E}[ I\!E_s^{n_1}I_s^{n_2-2}(I_s^{*})^{n_3}]ds\\
  &\color{blue}\quad + \frac{1}{2} n_2(n_2-1)\theta  \times&\color{blue}\int_0^t \mathbb{E}[ I\!E_s^{n_1}I_s^{n_2-2}(I_s^{*})^{n_3}]ds\\
  &\color{blue}\quad + \frac{1}{2} n_2(n_2-1)\sigma_v  \times&\color{blue}\int_0^t e^{-ks} \mathbb{E}[ I\!E_s^{n_1+1}I_s^{n_2-2}(I_s^{*})^{n_3}]ds\\
  &\color{blue}\quad + \frac{1}{2} n_2(n_2-1)  \times&\color{blue}\int_0^t e^{-ks}I\!E\!Z_s \mathbb{E}[ I\!E_s^{n_1}I_s^{n_2-2}(I_s^{*})^{n_3}]ds\\
  &\quad + n_1n_2(v_0-\theta) \times&\int_0^t \mathbb{E}[ I\!E_s^{n_1-1}I_s^{n_2-1}(I_s^{*})^{n_3}]ds\\
  &\quad + n_1n_2\theta         \times&\int_0^t e^{ks} \mathbb{E}[ I\!E_s^{n_1-1}I_s^{n_2-1}(I_s^{*})^{n_3}]ds\\
  &\quad + n_1n_2\sigma_v       \times&\int_0^t \mathbb{E}[ I\!E_s^{n_1}I_s^{n_2-1}(I_s^{*})^{n_3}]ds\\
  &\quad + n_1n_2                 \times&\int_0^t I\!E\!Z_s \mathbb{E}[ I\!E_s^{n_1-1}I_s^{n_2-1}(I_s^{*})^{n_3}]ds\\
  &\color{blue}\quad + \frac{1}{2}n_3(n_3-1)(v_0-\theta)  \times&\color{blue} \int_0^t e^{-ks} \mathbb{E}[ I\!E_s^{n_1} I_s^{n_2} (I_s^{*})^{n_3-2}]ds\\
  &\color{blue}\quad + \frac{1}{2}n_3(n_3-1)\theta  \times&\color{blue} \int_0^t  \mathbb{E}[ I\!E_s^{n_1} I_s^{n_2} (I_s^{*})^{n_3-2}]ds\\
  &\color{blue}\quad + \frac{1}{2}n_3(n_3-1)\sigma_v  \times&\color{blue} \int_0^t e^{-ks} \mathbb{E}[ I\!E_s^{n_1+1} I_s^{n_2} (I_s^{*})^{n_3-2}]ds\\
  &\color{blue}\quad + \frac{1}{2}n_3(n_3-1)  \times&\color{blue} \int_0^t e^{-ks} I\!E\!Z_s \mathbb{E}[ I\!E_s^{n_1} I_s^{n_2} (I_s^{*})^{n_3-2}]ds,
  \end{align*}

in which, for notation simplification, the condition notation
:math:`|v_0, z(s), 0\le s\le t` is also removed from
all conditional expectations on the right-hand side of above equation.

We decode
:math:`\mathbb{E}[I\!E_t^{n_1} I_t^{n_2} (I_t^{*})^{n_3}|v_0, z(s), 0\le s\le t]`
as the following :py:class:`~ajdmom.poly.Poly`,

.. math::

  \begin{align*}
  &\mathbb{E}[I\!E_t^{n_1} I_t^{n_2} (I_t^{*})^{n_3}|v_0, z(s), 0\le s\le t]\\
  &= \sum_{\boldsymbol{j}, \boldsymbol{l}, \boldsymbol{o}, \boldsymbol{p},
  \boldsymbol{q}}
  c_{\boldsymbol{j}, \boldsymbol{l}, \boldsymbol{o}, \boldsymbol{p},
  \boldsymbol{q}}
  e^{j_1kt} t^{j_2} k^{-j_3} (v_0-\theta)^{j_4} \theta^{j_5} \sigma_v^{j_6}
  f_{Z_t}(\boldsymbol{l}, \boldsymbol{o}, \boldsymbol{p},\boldsymbol{q}),
  \end{align*}

where vectors :math:`\boldsymbol{j}, \boldsymbol{l}, \boldsymbol{o},
\boldsymbol{p}, \boldsymbol{q}` denote

.. math::

   \begin{align*}
   \boldsymbol{j} \equiv (j_1, \cdots, j_6),\\
   \boldsymbol{l} \equiv (l_1, \cdots, l_n),\quad
   \boldsymbol{o} \equiv (o_1, \cdots, o_n),\\
   \boldsymbol{p} \equiv (p_2, \cdots, p_n),\quad
   \boldsymbol{q} \equiv (q_2, \cdots, q_n),
   \end{align*}

respectively,
:math:`c_{\boldsymbol{j}, \boldsymbol{l}, \boldsymbol{o}, \boldsymbol{p},\boldsymbol{q}}`
denotes the corresponding constant coefficient
and function :math:`f_{Z_t}(\boldsymbol{l}, \boldsymbol{o},
\boldsymbol{p}, \boldsymbol{q})` is defined as

.. math::
  :label: fZ

   \begin{align*}
   &f_{Z_t}(\boldsymbol{l},\boldsymbol{o},\boldsymbol{p},\boldsymbol{q})\\
   &\triangleq \sum_{i_1=1}^{N(t)}\cdots\sum_{i_n=1}^{N(t)}
     e^{l_1ks_{i_1} + \cdots + l_nks_{i_n}} J_{i_1} \cdots J_{i_n}
     s_{i_1}^{o_1}\cdots s_{i_n}^{o_n} \\
   &\qquad\qquad\qquad \cdot
    e^{p_2k(s_{i_1}\vee s_{i_2}) + \cdots + p_nk(s_{i_1}\vee \cdots
       \vee s_{i_n})}\\
   &\qquad\qquad\qquad \cdot
    (s_{i_1}\vee s_{i_2})^{q_2}\cdots
    (s_{i_1}\vee \cdots \vee s_{i_n})^{q_n}.
    \end{align*}

It should be noted that

- :math:`s_{i1}\vee s_{i2} \equiv \max\{s_{i1}, s_{i2}\}`,
  :math:`s_{i1}\vee s_{i2} \vee s_{i3} \equiv \max\{s_{i1}, s_{i2}, s_{i3}\}`,
  so on and so forth.

- when :math:`n=1`, :math:`\boldsymbol{p} = \boldsymbol{q} = ()`,
  :math:`f_{Z_t}(\boldsymbol{l}, \boldsymbol{o}, \boldsymbol{p},
  \boldsymbol{q}) = \sum_{i_1=1}^{N(t)} e^{l_1ks_{i_1}}J_{i_1} s_{i_1}^{o_1}`.

- when :math:`n=0`, :math:`\boldsymbol{l} = \boldsymbol{o} =
  \boldsymbol{p} = \boldsymbol{q} = ()`,
  :math:`f_{Z_t}(\boldsymbol{l}, \boldsymbol{o}, \boldsymbol{p},
  \boldsymbol{q}) = 1`.

Example formulas for :math:`n_1+n_2=2` combinations,

.. math::

  \begin{align*}
  &\mathbb{E}[I\!E_t^2 |v_0, z(s), 0\le s\le t] \\
  &= (v_0-\theta)k^{-1}(e^{kt}-1) + \frac{1}{2}\theta k^{-1} (e^{2kt} - 1)
  + k^{-1}\sum_{i=1}^{N(t)} e^{ks_i}J_i (e^{kt} - e^{ks_i}),\\
  &\mathbb{E}[I\!E_t I_t |v_0, z(s), 0\le s\le t] \\
  &= (v_0-\theta) t + \theta k^{-1} (e^{kt} - 1)
  + \sum_{i=1}^{N(t)} e^{ks_i}J_i (t-s_i),\\
  &\mathbb{E}[I_t^2 |v_0, z(s), 0\le s\le t] \\
  &= -(v_0-\theta)k^{-1}(e^{-kt} - 1) + \theta t
  - k^{-1}\sum_{i=1}^{N(t)} e^{ks_i}J_i (e^{-kt} - e^{-ks_i}).
  \end{align*}


Integrals
===========

The essential computation now becomes

.. math::
   :label: int_et_fZ

   \int_{0}^t e^{iks} s^j f_{Z_s}(\boldsymbol{l}, \boldsymbol{o},
    \boldsymbol{p}, \boldsymbol{q}) ds.

We present the result and implementation first.

.. math::

  \int_{0}^t e^{iks} s^j f_{Z_s}(\boldsymbol{l}, \boldsymbol{o},
    \boldsymbol{p}, \boldsymbol{q}) ds
  = F_{Z_t}(\boldsymbol{l}, \boldsymbol{o}, \boldsymbol{p}, \boldsymbol{q},
  i, j),

where the function on the right-hand side is defined as

.. math::

  \begin{align*}
   &F_{Z_t}(\boldsymbol{l},\boldsymbol{o},\boldsymbol{p},\boldsymbol{q},
    i, j)\\
   &\triangleq \sum_{i_1=1}^{N(t)}\cdots\sum_{i_n=1}^{N(t)}
     e^{l_1ks_{i_1} + \cdots + l_nks_{i_n}} J_{i_1} \cdots J_{i_n}
     s_{i_1}^{o_1}\cdots s_{i_n}^{o_n} \\
   &\qquad\qquad\qquad \cdot
    e^{p_2k(s_{i_1}\vee s_{i_2}) + \cdots + p_nk(s_{i_1}\vee \cdots
       \vee s_{i_n})}\\
   &\qquad\qquad\qquad \cdot
    (s_{i_1}\vee s_{i_2})^{q_2}\cdots
    (s_{i_1}\vee \cdots \vee s_{i_n})^{q_n}\\
   &\qquad\qquad\qquad \cdot
   \int_{s_{i_1}\vee \cdots \vee s_{i_n}}^t e^{iks} s^j ds.
   \end{align*}

The integral in :eq:`int_et_fZ` is implemented in
:py:func:`~ajdmom.ito_cond2_mom.int_et_fZ` in module
:py:mod:`~ajdmom.ito_cond2_mom`. The integral
:math:`\int_{s_{i_1}\vee \cdots \vee s_{i_n}}^t e^{iks} s^j ds` can be
calculated as we did for :math:`\int_0^t e^{iks} s^j ds` in
:doc:`../generated/ajdmom.ito_mom`.

Then we explain the calculations.
Let's take a look at a simple example,
:math:`\int_0^te^{ks}I\!E\!Z_s ds`.

.. math::

  e^{ks}I\!E\!Z_s =
   \begin{cases}
     0,                                     &0 ~~~~     \le s<s_1,\\
     e^{ks}\sum_{i=1}^1 e^{ks_i}J_i,        &s_1~~~     \le s<s_2,\\
                                                  &   \vdots     \\
     e^{ks}\sum_{i=1}^{N(t)-1} e^{ks_i}J_i, &s_{N(t)-1}  \le s<s_{N(t)},\\
     e^{ks}\sum_{i=1}^{N(t)} e^{ks_i}J_i,   &s_{N(t)}~~~     \le s< t.
   \end{cases}

.. math::

  \begin{align*}
  &\int_0^t e^{ks}I\!E\!Z_s ds\\
  &= \int_0^{s_1} e^{ks}I\!E\!Z_s ds + \int_{s_1}^{s_2} e^{ks}I\!E\!Z_s ds
    + \cdots + \int_{s_n}^{t} e^{ks}I\!E\!Z_s ds\\
  &= \frac{1}{k}(e^{ks_2} - e^{ks_1}) \sum_{i=1}^{1}e^{ks_i}J_i + \cdots
    + \frac{1}{k}(e^{ks_n} - e^{ks_{n-1}})\sum_{i=1}^{N(t)-1}e^{ks_i}J_i\\
  &  \quad + \frac{1}{k}(e^{kt} - e^{ks_n})\sum_{i=1}^{N(t)}e^{ks_i}J_i\\
  &= \sum_{i=1}^{N(t)} e^{ks_i}J_i \frac{1}{k}(e^{kt} - e^{ks_i}).
  \end{align*}

In short,

.. math::

  \begin{align*}
  I\!E\!Z_t &= \sum_{i=1}^{N(t)} e^{ks_i}J_i,\\
  \int_0^t e^{ks}I\!E\!Z_s ds &= \sum_{i=1}^{N(t)} e^{ks_i}J_i
  \frac{1}{k} (e^{kt} - e^{ks_i}).
  \end{align*}

Another example

.. math::

  \int_0^t e^{ks} I\!E\!Z_s I\!E\!Z_s ds = ?

.. math::

  \begin{align*}
  &\int_0^t e^{ks} \sum_{i=1}^{N(s)}\sum_{j=1}^{N(s)} e^{ks_i+ks_j}J_iJ_jds\\
  &= \sum_{i=1}^{N(t)}\sum_{j=1}^{N(t)}e^{ks_i+ks_j}J_iJ_j \frac{1}{k}
  (e^{kt} -e^{k(s_i\vee s_j)}).
  \end{align*}

The two examples should have explained the derivation well.

Implementation summary
-----------------------

1. Define :py:func:`~ajdmom.ito_cond2_mom.recursive_IEII` to realize the
   recursive step in equation :eq:`ito-jmp-moment-ii`.

2. Define :py:func:`~ajdmom.ito_cond2_mom.moment_IEII` to finish the computation
   of :math:`\mathbb{E}[I\!E_t^{n_1}I_t^{n_2}(I_t^{*})^{n_3}|v_0, z_s,
   0\le s\le t]`.

"""
import math
from fractions import Fraction as Frac

from ajdmom.poly import Poly
from ajdmom.ito_mom import c_nmi
from ajdmom.utils import fZ_old


def int_et_fZ(n, m, N_sum):
    r"""integral of :math:`\int_0^t e^{iks} s^j f_{Z_s}(l,o,p,q)ds`

    For each element with index :math:`(s_{i1},\dots,s_{in})`,
    the integral becomes
    :math:`\int_{s_{i1}\vee\cdots\vee s_{in}}^t e^{iks} s^j ds`.
    When there is no summation at all in :math:`f_{Z_s}(l,o,p,q)`,
    the whole integral simplifies to :math:`\int_0^t e^{iks} s^jds`.

    :param int n: power of :math:`e^{ks}`, i.e., :math:`i`.
    :param int m: power of :math:`s`, i.e., :math:`j`.
    :param int N_sum: level of summations in :math:`f_{Z_s}(l,o,p,q)`.
    :return: a poly with attribute ``keyfor`` =
      ('e^{kt}', 't', 'k^{-}', 'e^{k(s_i1 v...v s_in)}', '(s_i1 v...v s_in)').
    :rtype: Poly
    """
    if m < 0:
        msg = f"m in int_et_fZ(n,m) equals {m}, however it must be non-negative!"
        raise ValueError(msg)
    #
    poly = Poly()
    kf = ['e^{kt}', 't', 'k^{-}', 'e^{k(s_i1 v...v s_in)}', '(s_i1 v...v s_in)']
    poly.set_keyfor(kf)
    #
    if N_sum == 0:  # lower bound -> 0
        if n == 0:
            poly[(0, m + 1, 0, 0, 0)] = Frac(1, m + 1)
        elif n != 0 and m == 0:
            poly[(n, 0, 1, 0, 0)] = Frac(1, n)
            poly[(0, 0, 1, 0, 0)] = -Frac(1, n)  # - F(0), F(0) != 0
        else:
            poly[(n, m, 1, 0, 0)] = Frac(1, n)
            for i in range(1, m + 1):
                c = c_nmi(n, m, i)
                poly[(n, m - i, i + 1, 0, 0)] = c
                if i == m:  # - F(0): - c_nmi, F(0) != 0
                    poly[(0, 0, i + 1, 0, 0)] = -c
    else:  # lower bound -> (s_i1 v...v s_in)
        if n == 0:
            poly[(0, m + 1, 0, 0, 0)] = Frac(1, m + 1)
            poly[(0, 0, 0, 0, m + 1)] = -Frac(1, m + 1)
        elif n != 0 and m == 0:
            poly[(n, 0, 1, 0, 0)] = Frac(1, n)
            poly[(0, 0, 1, n, 0)] = -Frac(1, n)
        else:
            poly[(n, m, 1, 0, 0)] = Frac(1, n)
            poly[(0, 0, 1, n, m)] = -Frac(1, n)
            for i in range(1, m + 1):
                c = c_nmi(n, m, i)
                poly[(n, m - i, i + 1, 0, 0)] = c
                poly[(0, 0, i + 1, n, m - i)] = -c
    return poly


def int_e_poly(coef, tp, m, poly):
    r"""integral of :math:`coef \times tp \times \int_0^t e^{mks} poly ds`

    :param Fraction coef: coefficient to multiply with
    :param int tp: type of the multiplication,

       +----+----------------------------+
       | tp | multiply with              |
       +====+============================+
       | 1  | :math:`v_0-\theta`         |
       +----+----------------------------+
       | 2  | :math:`\theta`             |
       +----+----------------------------+
       | 3  | :math:`\sigma_v`           |
       +----+----------------------------+

    :param int m: power of :math:`e^{ks}` in the integrand
    :param Poly poly: a Poly object, such as
      :math:`\mathbb{E}[I\!E_t^{n_1} I_t^{n_2} I_t^{*n_3}
      |v_0, z(s), 0\le s\le t]`, with attribute ``keyfor`` =
      ('e^{kt}', 't', 'k^{-}', 'v0-theta', 'theta', 'sigma',
      'l_{1:n}', 'o_{1:n}', 'p_{2:n}', 'q_{2:n}')
    :return: a Poly with the same ``keyfor`` of the input poly
    :rtype: Poly
    """
    poln = Poly()
    poln.set_keyfor(poly.keyfor)  # poly = IEII[(n1, n2, n3)]
    for k in poly:
        poly_sub = int_et_fZ(m + k[0], k[1], len(k[6]))  # \int e^{iks} s^{j} ds
        # ['e^{kt}', 't', 'k^{-}', 'e^{k(s_i1 v...v s_in)}', '(s_i1 v...v s_in)']
        for kk in poly_sub:
            # key1 + key2
            if tp == 1:
                key1 = (kk[0], kk[1], kk[2] + k[2], k[3] + 1, k[4], k[5])
            elif tp == 2:
                key1 = (kk[0], kk[1], kk[2] + k[2], k[3], k[4] + 1, k[5])
            else:
                key1 = (kk[0], kk[1], kk[2] + k[2], k[3], k[4], k[5] + 1)
            #
            if len(k[6]) == 0:  # 0 summation  inside
                key2 = ((), (), (), ())
            elif len(k[6]) == 1:  # 1 summation  inside
                k6 = (k[6][0] + kk[3],)
                k7 = (k[7][0] + kk[4],)
                key2 = (k6, k7, (), ())
            else:  # 2 summations inside or more
                k8 = list(k[8])
                k8[-1] += kk[3]
                k8 = tuple(k8)
                k9 = list(k[9])
                k9[-1] += kk[4]
                k9 = tuple(k9)
                key2 = (k[6], k[7], k8, k9)
            poln.add_keyval(key1 + key2, coef * poly[k] * poly_sub[kk])
    return poln


def int_e_IEZ_poly(coef, m, poly):
    r"""integral of :math:`coef \times \int_0^t e^{mks} I\!E\!Z_s poly ds`

    :param Fraction coef: coefficient to multiply with
    :param int m: power of :math:`e^{ks}` in the integrand
    :param Poly poly: a Poly object, such as
      :math:`\mathbb{E}[I\!E_t^{n_1} I_t^{n_2} I_t^{*n_3}
      |v_0, z(s), 0\le s\le t]`, with attribute ``keyfor`` =
      ('e^{kt}', 't', 'k^{-}', 'v0-theta', 'theta', 'sigma',
      'l_{1:n}', 'o_{1:n}', 'p_{2:n}', 'q_{2:n}')
    :return: a Poly with the same ``keyfor`` of the input poly
    :rtype: Poly
    """
    poln = Poly()
    poln.set_keyfor(poly.keyfor)  # poly = IEII[(n1, n2, n3)]
    for k in poly:
        # IEZ times poly
        k6 = k[6] + (1,)
        k7 = k[7] + (0,)
        if len(k6) >= 2:
            k8 = k[8] + (0,)
            k9 = k[9] + (0,)
        #
        # \int e^{iks} s^{j} poly ds
        poly_sub = int_et_fZ(m + k[0], k[1], len(k6))
        # ['e^{kt}', 't', 'k^{-}', 'e^{k(s_i1 v...v s_in)}', '(s_i1 v...v s_in)']
        for kk in poly_sub:
            # key1
            key1 = (kk[0], kk[1], kk[2] + k[2], k[3], k[4], k[5])
            # key2
            if len(k6) == 1:  # 1 summation  inside
                # k6 = (k6[0] + kk[3],)
                # k7 = (k7[0] + kk[4],)
                # key2 = (k6, k7, (), ())
                key2 = ((k6[0] + kk[3],), (k7[0] + kk[4],), (), ())
            else:  # 2 summations inside or more
                # k8 = list(k8)
                # k8[-1] += kk[3]
                # k8 = tuple(k8)
                # k9 = list(k9)
                # k9[-1] += kk[4]
                # k9 = tuple(k9)
                # key2 = (k6, k7, k8, k9)
                key2 = (k6, k7, k8[0:-1] + (k8[-1] + kk[3],), k9[0:-1] + (k9[-1] + kk[4],))
            poln.add_keyval(key1 + key2, coef * poly[k] * poly_sub[kk])
    return poln


def recursive_IEII(n1, n2, n3, IEII):
    r"""Recursive step in equation :eq:`ito-jmp-moment-ii`

    :param int n1: power of :math:`I\!E_s` in the integrand.
    :param int n2: power of :math:`I_s` in the integrand.
    :param int n3: power of :math:`I_s^{*}` in the integrand.
    :param dict IEII: a dict of joint conditional moments of
      :math:`I\!E_sI_sI_s^{*}`, with key (n1, n2, n3) and value Poly object
      with attribute ``keyfor`` =
      ('e^{kt}','t','k^{-}','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}', 'p_{2:n}', 'q_{2:n}').
    :return: poly with the same ``keyfor`` of that of in IEII
    :rtype: Poly
    """
    kf = ['e^{kt}', 't', 'k^{-}', 'v0-theta', 'theta', 'sigma']
    kf += ['l_{1:n}', 'o_{1:n}', 'p_{2:n}', 'q_{2:n}']
    poly = Poly()
    poly.set_keyfor(kf)
    #
    if n1 >= 2 and n2 >= 0 and n3 >= 0:
        c = Frac(n1 * (n1 - 1), 2)
        poly.merge(int_e_poly(c, 1, 1, IEII[(n1 - 2, n2, n3)]))
        poly.merge(int_e_poly(c, 2, 2, IEII[(n1 - 2, n2, n3)]))
        poly.merge(int_e_poly(c, 3, 1, IEII[(n1 - 1, n2, n3)]))
        poly.merge(int_e_IEZ_poly(c, 1, IEII[(n1 - 2, n2, n3)]))
    if n1 >= 0 and n2 >= 2 and n3 >= 0:
        c = Frac(n2 * (n2 - 1), 2)
        poly.merge(int_e_poly(c, 1, -1, IEII[(n1, n2 - 2, n3)]))
        poly.merge(int_e_poly(c, 2, 0, IEII[(n1, n2 - 2, n3)]))
        poly.merge(int_e_poly(c, 3, -1, IEII[(n1 + 1, n2 - 2, n3)]))
        poly.merge(int_e_IEZ_poly(c, -1, IEII[(n1, n2 - 2, n3)]))
    if n1 >= 1 and n2 >= 1 and n3 >= 0:
        c = Frac(n1 * n2, 1)
        poly.merge(int_e_poly(c, 1, 0, IEII[(n1 - 1, n2 - 1, n3)]))
        poly.merge(int_e_poly(c, 2, 1, IEII[(n1 - 1, n2 - 1, n3)]))
        poly.merge(int_e_poly(c, 3, 0, IEII[(n1, n2 - 1, n3)]))
        poly.merge(int_e_IEZ_poly(c, 0, IEII[(n1 - 1, n2 - 1, n3)]))
    if n1 >= 0 and n2 >= 0 and n3 >= 2:
        c = Frac(n3 * (n3 - 1), 2)
        poly.merge(int_e_poly(c, 1, -1, IEII[(n1, n2, n3 - 2)]))
        poly.merge(int_e_poly(c, 2, 0, IEII[(n1, n2, n3 - 2)]))
        poly.merge(int_e_poly(c, 3, -1, IEII[(n1 + 1, n2, n3 - 2)]))
        poly.merge(int_e_IEZ_poly(c, -1, IEII[(n1, n2, n3 - 2)]))
    return poly


def moment_IEII(n1, n2, n3):
    r""":math:`\mathbb{E}[I\!E_t^{n_1}I_t^{n_2}(I_t^{*})^{n_3}
    |v_0, z(s), 0\le s\le t]`

    :param int n1: power of :math:`I\!E_t`.
    :param int n2: power of :math:`I_t`.
    :param int n3: power of :math:`I_t^{*}`.
    :return: poly with ``keyfor`` =
       ('e^{kt}','t','k^{-}','v0-theta','theta','sigma',
       'l_{1:n}', 'o_{1:n}', 'p_{2:n}', 'q_{2:n}').
    :rtype: Poly
    """
    if n1 + n2 + n3 < 0:
        raise ValueError(f"moment_IEII({n1},{n2},{n3}) is called!")
    #
    # IEII: a dict of conditional moments of E[IE_t^{n1}I_t^{n2}I_t^{*n3}]
    #
    IEII = {}
    #
    kf = ['e^{kt}', 't', 'k^{-}', 'v0-theta', 'theta', 'sigma']
    kf += ['l_{1:n}', 'o_{1:n}', 'p_{2:n}', 'q_{2:n}']
    #
    # special poly constants, analog to 0 and 1
    #
    P0 = Poly({(0, 0, 0, 0, 0, 0, (), (), (), ()): Frac(0, 1)})
    P0.set_keyfor(kf)
    P1 = Poly({(0, 0, 0, 0, 0, 0, (), (), (), ()): Frac(1, 1)})
    P1.set_keyfor(kf)
    #
    # n1 + n2 + n3 = 0: special case
    #
    IEII[(0, 0, 0)] = P1  # equiv to constant 1
    #
    # n1 + n2 + n3 = 1
    #
    IEII[(1, 0, 0)] = P0  # equiv to constant 0
    IEII[(0, 1, 0)] = P0
    IEII[(0, 0, 1)] = P0
    #
    # n1 + n2 +n3 = 2
    #
    P200 = Poly({
        (1, 0, 1, 1, 0, 0, (), (), (), ()): Frac(1, 1),
        (0, 0, 1, 1, 0, 0, (), (), (), ()): -Frac(1, 1),
        (2, 0, 1, 0, 1, 0, (), (), (), ()): Frac(1, 2),
        (0, 0, 1, 0, 1, 0, (), (), (), ()): -Frac(1, 2),
        (1, 0, 1, 0, 0, 0, (1,), (0,), (), ()): Frac(1, 1),
        (0, 0, 1, 0, 0, 0, (2,), (0,), (), ()): -Frac(1, 1)
    })
    P200.set_keyfor(kf)
    P110 = Poly({
        (0, 1, 0, 1, 0, 0, (), (), (), ()): Frac(1, 1),
        (1, 0, 1, 0, 1, 0, (), (), (), ()): Frac(1, 1),
        (0, 0, 1, 0, 1, 0, (), (), (), ()): -Frac(1, 1),
        (0, 1, 0, 0, 0, 0, (1,), (0,), (), ()): Frac(1, 1),
        (0, 0, 0, 0, 0, 0, (1,), (1,), (), ()): -Frac(1, 1)
    })
    P110.set_keyfor(kf)
    P020 = Poly({
        (-1, 0, 1, 1, 0, 0, (), (), (), ()): -Frac(1, 1),
        (0, 0, 1, 1, 0, 0, (), (), (), ()): Frac(1, 1),
        (0, 1, 0, 0, 1, 0, (), (), (), ()): Frac(1, 1),
        (-1, 0, 1, 0, 0, 0, (1,), (0,), (), ()): -Frac(1, 1),
        (0, 0, 1, 0, 0, 0, (0,), (0,), (), ()): Frac(1, 1)
    })
    P020.set_keyfor(kf)
    IEII[(2, 0, 0)] = P200
    IEII[(1, 1, 0)] = P110
    IEII[(1, 0, 1)] = P0
    IEII[(0, 2, 0)] = P020
    IEII[(0, 1, 1)] = P0
    IEII[(0, 0, 2)] = P020
    #
    if n1 + n2 + n3 <= 2: return IEII[(n1, n2, n3)]
    #
    # n1 + n2 + n3 >= 3: typical cases
    #
    if n1 + n2 + n3 > 3:
        # compute all lower-order moments to get ready
        for n in range(3, n1 + n2 + n3):
            for i in range(n, -1, -1):
                for j in range(n - i, -1, -1):
                    poly = recursive_IEII(i, j, n - i - j, IEII)
                    poly.remove_zero()
                    IEII[(i, j, n - i - j)] = poly
            # delete polys no more needed
            index = [key for key in IEII if key[0] + key[1] + key[2] == n - 2]
            for key in index: del IEII[key]
    # the last one
    poly = recursive_IEII(n1, n2, n3, IEII)
    poly.remove_zero()
    return poly


def poly2num(poly, par):
    """Convert a polynomial to numerical value

    :param Poly poly: polynomial to convert, with ``keyfor`` =
       ('e^{kt}','t','k^{-}','v0-theta','theta','sigma',
       'l_{1:n}', 'o_{1:n}', 'p_{2:n}', 'q_{2:n}').
    :param dict par: parameters, including jumpsize and jumptime.
    :return: numerical value.
    :rtype: float
    """
    k, h, v0, theta, sigma = par['k'], par['h'], par['v0'], par['theta'], par['sigma']
    J = par['jumpsize'] # vector of jump sizes
    s = par['jumptime'] # vector of jump time points
    f = 0
    for K in poly:
        val = math.exp(K[0] * k * h) * (h ** K[1]) * (k ** (-K[2]))
        val *= ((v0 - theta) ** K[3]) * (theta ** K[4]) * (sigma ** K[5])
        l, o, p, q = K[6], K[7], K[8], K[9]
        val *= fZ_old(l, o, p, q, k, s, J)
        f += val * poly[K]
    return f


if __name__ == "__main__":
    import sys
    from pprint import pprint

    print('\nExample usage of the module function\n')
    kf = ['e^{kt}', 't', 'k^{-}', 'v0-theta', 'theta', 'sigma']
    kf += ['l_{1:n}', 'o_{1:n}', 'p_{2:n}', 'q_{2:n}']
    print(f"moment_IEII(n1, n2, n3) returns a poly with keyfor = \n{kf}")
    #
    args = sys.argv[1:]
    n = 3 if len(args) == 0 else int(args[0])
    for n1 in range(n, -1, -1):
        for n2 in range(n - n1, -1, -1):
            n3 = n - n1 - n2
            print(f'\nmoment_IEII({n1},{n2},{n3}) = ')
            pprint(moment_IEII(n1, n2, n3))



"""
Module for generating a trajectory of samples from mdl_2fsvj by Euler approximation

To facilitate the verification of the correctness of our codes
by comparison the population moments (given by our package) and
their sample counterparts.
Here I define function :py:func:`~ajdmom.mdl_2fsvj.euler.r2FSVJ`.
"""
import math
import numpy as np


def r2FSVJ(v0, par, N, n_segment=10, h=1):
    """Generate a trajectory of samples from mdl_2fsvj by Euler approximation

    :param float v0: values of the initial variances :math:`(v_1(0),v_2(0)) > (0,0)`.
    :param dict par: parameters in a dict.
    :param int N: target length of samples to generate.
    :param int n_segment: number of segments each interval is splitted into.
    :param float h: time interval between each two consecutive samples.
    :return: a sequence of samples.
    :rtype: numpy 1-D array.
    """
    mu = par['mu']
    k1 = par['k1']
    theta1 = par['theta1']
    sigma1 = par['sigma_v1']
    k2 = par['k2']
    theta2 = par['theta2']
    sigma2 = par['sigma_v2']
    lmbd = par['lambda']
    mu_j = par['mu_j']
    sigma_j = par['sigma_j']
    #
    dlt = h / n_segment
    std = math.sqrt(dlt)
    #
    y_series = np.empty(N)
    #
    v1_0 = v0[0]
    v2_0 = v0[1]
    #
    for i in range(N):
        I = 0
        IV1 = 0
        IV2 = 0
        for j in range(n_segment):
            eps = math.sqrt(v1_0 + v2_0) * np.random.normal(0, std)
            eps1 = math.sqrt(v1_0) * np.random.normal(0, std)
            eps2 = math.sqrt(v2_0) * np.random.normal(0, std)
            v1 = v1_0 + k1 * (theta1 - v1_0) * dlt + sigma1 * eps1
            v2 = v2_0 + k2 * (theta2 - v2_0) * dlt + sigma2 * eps2
            #
            if v1 < 0:
                v1 = 0  # eps1 = -(v1_0 + k1*(theta1-v1_0)*dlt)/sigma1
            if v2 < 0:
                v2 = 0  # eps2 = -(v2_0 + k2*(theta2-v2_0)*dlt)/sigma2
            #
            IV1 += v1_0 * dlt
            IV2 += v2_0 * dlt
            I += eps
            v1_0 = v1
            v2_0 = v2
        #
        # jump part
        n = np.random.poisson(lmbd * h)
        J_n = 0
        if n > 0:
            for l in range(n):
                J_n += np.random.normal(mu_j, sigma_j)
        #
        y = mu * h - (IV1 + IV2) / 2 + I + J_n
        y_series[i] = y
    return y_series



"""
Conditional Moments for the SVIJ model, given :math:`v_0`
and the realized jumps in the variance.
"""
import math
from fractions import Fraction as Frac

from ajdmom.poly import Poly
from ajdmom.cpp_mom import mcpp
from ajdmom.utils import fZ
from ajdmom.mdl_svvj.cond2_mom import moments_y_to as moments_y_to_svvj


def moments_y_to(l):
    r"""conditional moments of :math:`y_t` of orders :math:`0:l`

    :param int l: highest order of the conditional moments to derive, >= 1
    :return: a list of polys with attribute ``keyfor`` =
      ('e^{kt}','t','k^{-}','beta_t','mu-theta/2','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}', 'p',
      'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)',
      'lmbd_s', 'mu_s', 'sigma_s'), now 'p' encoding power of :math:`I\!Z_t`.
    :rtype: Poly
    """
    moms = []
    moms_svvj = moments_y_to_svvj(l)  # 0-th to l-th moments
    #
    kf = moms_svvj[0].keyfor + ('lmbd_s', 'mu_s', 'sigma_s')
    #
    # special case: 0-th moment
    #
    P1 = Poly({(0, 0, 0, 0, 0, 0, 0, 0, (), (), 0, 0, 0, 0, 0, 0, 0): Frac(1, 1)})
    P1.set_keyfor(kf)
    moms.append(P1)  # equiv to constant 1
    #
    # typical cases
    #
    for n in range(1, l + 1):
        poly = Poly()
        poly.set_keyfor(kf)
        for n1 in range(n, -1, -1):
            c = math.comb(n, n1)
            pol1 = moms_svvj[n1]  # 0-th moment supported
            pol2 = mcpp(n - n1)  # kf = ['lmbd_s*h','mu_s','sigma_s']
            for k1 in pol1:
                for k2 in pol2:
                    key = (k1[0], k1[1] + k2[0]) + k1[2:] + k2  # t = h
                    poly.add_keyval(key, c * pol1[k1] * pol2[k2])
        poly.remove_zero()
        moms.append(poly)
    return moms


def moment_y(l):
    moms = moments_y_to(l)
    return moms[-1]


##########
# scalar
##########

def poly2num(poly, par):
    """Decode poly back to scalar

    :param Poly poly: poly to be decoded with attribute ``keyfor`` =
      ('e^{kt}','t','k^{-}','beta_t','mu-theta/2','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}', 'p',
      'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)',
      'lmbd_s', 'mu_s', 'sigma_s')
    :param dict par: parameters in dict, ``jumptime`` (tuple) and ``jumpsize`` (tuple)
      must also be included

    :return: scalar of the poly
    :rtype: float
    """
    v0 = par['v0']
    mu = par['mu']
    k = par['k']
    theta = par['theta']
    sigma = par['sigma']
    rho = par['rho']
    h = par['h']  # t= h
    J = par['jumpsize']  # vector of jump sizes
    s = par['jumptime']  # vector of jump time points
    lmbd_s = par['lmbd_s']
    mu_s = par['mu_s']
    sigma_s = par['sigma_s']
    #
    beta_t = (1 - math.exp(-k * h)) / (2 * k)
    #
    f = 0
    for K in poly:
        val = math.exp(K[0] * k * h) * (h ** K[1]) * (k ** (-K[2]))
        val *= (beta_t ** K[3]) * ((mu - theta / 2) ** K[4]) * ((v0 - theta) ** K[5])
        val *= (theta ** K[6]) * (sigma ** K[7])
        #
        l, o, p = K[8], K[9], K[10]
        #
        val *= fZ(l, o, k, s, J)
        #
        IZ = sum(J)
        val *= IZ ** p
        #
        val *= ((sigma / (2 * k)) ** K[11]) * ((rho - sigma / (2 * k)) ** K[12])
        val *= ((1 - rho ** 2) ** (K[13] / 2))
        #
        val *= (lmbd_s ** K[14]) * (mu_s ** K[15]) * (sigma_s ** K[16])
        #
        f += val * poly[K]
    return f


def m(l, par):
    """conditional moment in scalar

    :param int l: order of the conditional moment
    :param dict par: parameters in dict, ``jumptime`` (tuple) and ``jumpsize`` (tuple)
      must also be included
    :return: scalar of the moment
    :rtype: float
    """
    moment = moment_y(l)
    value = poly2num(moment, par)
    return value


if __name__ == "__main__":
    # Example usage of the module, see 'tests/test_mdl_svcj.py' for more test
    from pprint import pprint
    import sys

    print('Example usage of the module function\n')
    args = sys.argv[1:]
    l = 1 if len(args) == 0 else int(args[0])
    moms = moments_y_to(l)  # 0-th to l-th moments
    kf = moms[0].keyfor
    print(f"moment_y({l}) = \n")
    pprint(moms[l])
    print(f"\nwhich is a poly with keyfor = \n{kf}")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys
import os

sys.path.insert(0, os.path.abspath('../../src'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'ajdmom'
copyright = '2023-2025, Yan-Feng Wu'
author = 'Yan-Feng Wu'
release = '3.1'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.duration',
              'sphinx.ext.doctest',
              'sphinx.ext.autodoc',
              'sphinx.ext.autosummary',
              'sphinx.ext.intersphinx',
              'sphinx.ext.mathjax',
              'sphinx.ext.viewcode',]
intersphinx_mapping = {
  'python': ('https://docs.python.org/', None)
}

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
# html_static_path = ['_static']



"""
Conditional moments (II)

Conditional moments of the SRJD model, given the initial state
and the realized jumps over the concerned interval.
"""
import math
from fractions import Fraction as Frac

from ajdmom.poly import Poly
from ajdmom.utils import comb


def int_e_poly(i, poly):
    r"""integral of :math:`\int_0^t e^{iks} poly ds`

    :param int i: power of :math:`e^{ks}`
    :param Poly poly: poly attribute ``keyfor`` =
      ('v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}')
    :return: poly with the same attribute ``keyfor`` of the input poly
    :rtype: Poly
    """
    poln = Poly()
    kf = ['v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}']
    poln.set_keyfor(kf)
    #
    for k in poly:
        # same 1st part of the integration
        key = (k[0], k[1] + 1, k[2], k[3], k[4] + i, k[5])
        val = poly[k] * Frac(1, k[4] + i)
        poln.add_keyval(key, val)
        # diff 2nd part of the integration
        k5 = key[5]
        if len(k5) == 0: # no summation inside: - e^{(k[4]+i)*k*0}
            key = key[0:4] + (0, ())
        else: # 1 or more  summations inside: - e^{(k[4]+i)*k*(s_i1 v ... v s_in)}
            k5_new = k[5][0:-1] + (k[5][-1] + k[4] + i, )
            key = key[0:4] + (0, k5_new)
        poln.add_keyval(key, -val)
    return poln


def int_e_IE(i, m, IE):
    r"""Integral of :math:`\int_0^t e^{iks} I\!E_s^m ds`

    :param int i: power of :math:`e^{ks}`.
    :param int m: power of :math:`IE_s`.
    :param dict IE: dict with key m and value Poly object with attribute ``keyfor`` =
      ('v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}')
    :return: poly with the same attribute ``keyfor`` of the Poly object in IE
    :rtype: Poly
    """
    return int_e_poly(i, IE[m])


def int_e_IEZ_IE(m, IE):
    r"""Integral of :math:`\int_0^t e^{ks} I\!E\!Z_s I\!E_s^m ds`

    :param int m: power of :math:`I\!E_s`
    :param dict IE: dict with key m and value Poly object with attribute ``keyfor`` =
      ('v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}')
    :return: poly with the same attribute ``keyfor`` of the Poly object in IE
    :rtype: Poly
    """
    b = IE[m]
    poly = Poly()
    kf = ['v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}']
    poly.set_keyfor(kf)
    #
    # IEZ times IE_s^m
    #
    for k in b:
        key = list(k)
        key[5] = k[5] + (0,)  # extend
        key = tuple(key)
        poly.add_keyval(key, b[k])
    #
    # \int_0^t e^{ks} IEZ_s IE_s^m ds
    #
    return int_e_poly(1, poly)


def coef_poly(coef, poly, tp):
    """Multiply poly with different type coefficients

    :param fractions.Fraction coef: coefficient.
    :param Poly poly: poly with attribute ``keyfor`` =
      ('v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}')
    :param int tp: type of the multiplication,

       +----+----------------------------+
       | tp | multiply with              |
       +====+============================+
       | 1  | :math:`v_0-\\theta`         |
       +----+----------------------------+
       | 2  | :math:`\\theta`             |
       +----+----------------------------+
       | 3  | :math:`1`                  |
       +----+----------------------------+
       | 4  | :math:`\\sigma_v`           |
       +----+----------------------------+

    :return: poly with the same attribute ``keyfor`` of the input poly
    :rtype: Poly
    """
    poln = Poly()
    kf = ['v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}']
    poln.set_keyfor(kf)
    #
    if tp == 1:
        for k in poly:
            key = (k[0] + 1, k[1], k[2], k[3], k[4], k[5])
            val = coef * poly[k]
            poln.add_keyval(key, val)
            #
            key = (k[0], k[1], k[2] + 1, k[3], k[4], k[5])
            poln.add_keyval(key, -val)
    if tp == 2:
        for k in poly:
            key = (k[0], k[1], k[2] + 1, k[3], k[4], k[5])
            val = coef * poly[k]
            poln.add_keyval(key, val)
    if tp == 3:
        for k in poly:
            key = k
            val = coef * poly[k]
            poln.add_keyval(key, val)
    if tp == 4:
        for k in poly:
            key = (k[0], k[1], k[2], k[3] + 1, k[4], k[5])
            val = coef * poly[k]
            poln.add_keyval(key, val)
    return poln


def recursive_IE(m, IE):
    r"""Recursive step in equation :eq:`srjd-IE-moment`

    :param int m: power of :math:`I\!E_s` in the integrand.
    :param dict IE: a dict of conditional moments of :math:`I\!E_t`,
      with key ``m`` and value Poly object with attribute ``keyfor`` =
      ('v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}')
    :return: poly with the same attribute ``keyfor`` of the Poly object in ``IE``
    :rtype: Poly
    """
    poly = Poly()
    kf = ['v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}']
    poly.set_keyfor(kf)
    #
    if m >= 2:
        c = Frac(m * (m - 1), 2)
        poly.merge(coef_poly(c, int_e_IE(1, m - 2, IE), 1))
        poly.merge(coef_poly(c, int_e_IE(2, m - 2, IE), 2))
        poly.merge(coef_poly(c, int_e_IEZ_IE(m - 2, IE), 3))
        poly.merge(coef_poly(c, int_e_IE(1, m - 1, IE), 4))
    return poly


def moment_IE(m):
    r""":math:`\mathbb{E}[I\!E_t^m|v_0, z(s), 0\le s\le t]`

    :param int m: order of the conditional moment
    :return: poly if return_all=False else IE, a dict of conditional moments
      of :math:`I\!E_t`, with key ``m`` and value Poly object with attribute
      ``keyfor`` = ('v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}')
    :rtype: Poly or dict of Poly
    """
    # IE： dict of E[IE^{m}]
    IE = {}
    #
    kf = ['v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}']
    #
    P1 = Poly({(0, 0, 0, 0, 0, ()): Frac(1, 1)})  # equiv to 1
    P1.set_keyfor(kf)
    P0 = Poly({(0, 0, 0, 0, 0, ()): Frac(0, 1)})  # equiv to 0
    P0.set_keyfor(kf)
    # m = 0
    # support for special case
    IE[0] = P1
    # m = 1
    IE[1] = P0
    # m = 2
    poly = Poly({
        (0, 1, 1, 0, 2, ()): Frac(1, 2),
        (1, 1, 0, 0, 1, ()): Frac(1, 1),
        (0, 1, 1, 0, 1, ()): -Frac(1, 1),
        (0, 1, 1, 0, 0, ()): Frac(1, 2),
        (1, 1, 0, 0, 0, ()): -Frac(1, 1),
        (0, 1, 0, 0, 1, (0,)): Frac(1, 1),
        (0, 1, 0, 0, 0, (1,)): -Frac(1, 1)
    })
    poly.set_keyfor(kf)
    IE[2] = poly
    #
    if m <= 2:
        return IE[m]
    #
    if m > 3:
        # compute all lower-order moments to get ready
        for n in range(3, m):
            poly = recursive_IE(n, IE)
            poly.remove_zero()
            IE[n] = poly
            # delete polys no more needed
            index = [key for key in IE if key == n - 2]
            for key in index: del IE[key]
    # the last one
    poly = recursive_IE(m, IE)
    poly.remove_zero()
    return poly


def moment_v(m):
    r"""conditional moment :math:`\mathbb{E}[v_t^m|v_0, z(s), 0\le s\le t]`

    :param int m: order of the conditional moment of :math:`v(t)`
    :return: poly with attribute ``keyfor`` =
      ('v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}')
    :rtype: Poly
    """
    kf = ['v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}']
    poly = Poly()
    poly.set_keyfor(kf)
    #
    if m == 0:
        poly.add_keyval((0, 0, 0, 0, 0, ()), Frac(1, 1))
        return poly
    if m == 1:
        poly.add_keyval((1, 0, 0, 0, -1, ()), Frac(1, 1))
        poly.add_keyval((0, 0, 1, 0, -1, ()), -Frac(1, 1))
        poly.add_keyval((0, 0, 1, 0, 0, ()), Frac(1, 1))
        poly.add_keyval((0, 0, 0, 0, -1, (0,)), Frac(1, 1))
        return poly
    # m >= 2
    for j in range(m + 1):  # be careful of 0:m <- range(m+1)
        C1 = math.comb(m, j)
        # prepare pol1
        pol1 = Poly()
        pol1.set_keyfor(kf)
        for j1 in range(j, -1, -1):
            for j2 in range(j - j1, -1, -1):
                for j3 in range(j - j1 - j2, -1, -1):
                    j4 = j - j1 - j2 - j3
                    k5 = tuple(0 for _ in range(j4))
                    key = (j1, 0, j2 + j3, m - j, j3 - m, k5)
                    C2 = comb(j, [j1, j2, j3, j4])
                    val = C1 * C2 * ((-1) ** j2)
                    pol1.add_keyval(key, val)
        # prepare pol2
        pol2 = moment_IE(m - j)
        # pol1 * pol2
        for k1 in pol1:
            for k2 in pol2:  # works either m-j = 0 or m-j = 1
                key = (k2[0] + k1[0],
                       k2[1] + k1[1],
                       k2[2] + k1[2],
                       k2[3] + k1[3],
                       k2[4] + k1[4],
                       k2[5] + k1[5]) # k2[5] must be put in front of k1[5]
                val = pol1[k1] * pol2[k2]
                poly.add_keyval(key, val)
    poly.remove_zero()
    return poly


##########
# scalar
##########
def fZ(l, k, s, J):
    r"""function :math:`f_{Z_t}(l_{1:n})` as defined in :eq:`fZ_IE`

    :param tuple l: vector :math:`\boldsymbol{l}`, should be a tuple
    :param float k: parameter :math:`k`
    :param tuple s: jump time points
    :param tuple J: jump sizes
    :return: function value
    :rtype: float
    """
    # 1. Continuous part
    if len(l) == 0: return 1
    #
    # 2. Jump part, len(l) >= 1, may be single|mutiple summation
    # 2.1 no jump occurs
    if len(J) == 0: return 0
    # 2.2 >=1 jumps occur
    val = 0
    stack = []
    nextstep = "forward"  # downward|backward
    while len(stack) < len(l):
        if nextstep == "forward":
            if len(stack) < len(l) - 1:
                stack.append(0)
                nextstep = "forward"
            else:
                # come to the innermost-level summation
                for i in range(len(J)):  # index start with 0
                    stack.append(i)
                    index = stack.copy()
                    stack.pop()
                    # evaluate and add to sum
                    f = 1
                    s_max = s[index[0]]
                    for ii in range(len(index)):
                        j = index[ii]
                        s_j = s[j]
                        f *= math.exp(k * s_j) * J[j]
                        s_max = s_max if s_max > s_j else s_j
                        f *= math.exp(l[ii] * k * s_max)
                    val += f
                    #
                # checking whether all indexes traversed
                all_N = True
                for i in range(len(l) - 1):
                    if stack[i] != len(J) - 1:
                        all_N = False
                        break
                if all_N: break
                #
                nextstep = "backward"
        else:  # nextstep = "backward"
            i = stack.pop()
            if i == len(J) - 1:
                nextstep = "backward"
            else:
                stack.append(i + 1)
                nextstep = "forward"
    return val


def poly2num(poly, par):
    """Decode poly back to scalar

    :param Poly poly: poly to be decoded with attribute ``keyfor`` =
      ('v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}')
    :param dict par: parameters in dict, ``jumptime`` and ``jumpsize`` must also be
      included
    :return: scalar of the poly
    :rtype: float
    """
    v0 = par['v0']
    k = par['k']
    theta = par['theta']
    sigma = par['sigma']
    h = par['h']  # t = h
    J = par['jumpsize']  # vector of jump sizes
    s = par['jumptime']  # vector of jump time points
    #
    f = 0
    for K in poly:
        val = (v0 ** K[0]) * (k ** (-K[1])) * (theta ** K[2]) * (sigma ** K[3])
        val *= math.exp(K[4] * k * h)
        #
        l= K[5]
        val *= fZ(l, k, s, J)
        #
        f += val * poly[K]
    return f


def m(l, par):
    """conditional moment in scalar

    :param int l: order of the conditional moment
    :param dict par: parameters in dict, ``jumptime`` and ``jumpsize`` must also be
      included
    :return: scalar of the central moment
    :rtype: float
    """
    moment = moment_v(l)
    value = poly2num(moment, par)
    return value


if __name__ == "__main__":
    from pprint import pprint
    import sys

    print('\nExample usage of the module function\n')
    args = sys.argv[1:]
    n = 2 if len(args) == 0 else int(args[0])
    mom = moment_v(n)
    print(f"moment_v({n}) = ")
    pprint(mom)
    print(f"which is a Poly with attribute keyfor = \n{mom.keyfor}")
    #
    m = 3 if len(args) <= 1 else int(args[1])
    moment = moment_IE(m)
    print(f"\nmoment_IE({m}) = ")
    pprint(moment)
    print(f"which is a Poly with attribute keyfor = \n{moment.keyfor}")



r"""
Itô process moments under Single Square-Root Diffusion Process

The content has also been explained in :doc:`../design` page.

Insights
=========

All :math:`\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]`
can be represented as a "Polynomial" of the following form

.. _polynomial-representation:

.. math::

   &\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]\\
   &= \sum_{n_3,i,j,l,o,p,q} b_{n_3ijlopq} e^{n_3k(n-1)h} e^{ik[t-(n-1)h]}
   [t-(n-1)h]^jv_{n-1}^l k^{-o}\theta^p\sigma_v^q

where :math:`b_{ijlopq}` is the coefficient.

To facilitate the representation and corresponding operations, I designed
a new *class* :py:class:`~ajdmom.poly.Poly` which is derived from
:class:`~collections.UserDict` in the Python Standard Library
`collections <https://docs.python.org/3/library/collections.html>`_.

Integrals
==========

The essential computation in recursive equation :eq:`ito-moment`
of :doc:`../theory` is that of the following integral

.. math::

   \int_{(n-1)h}^t e^{ik[s-(n-1)h]} [s-(n-1)h]^j ds.


For the indefinite integral, we have

.. math::

   \int e^{nkt} t^m dt =
   \begin{cases}
   \sum_{i=0}^m c_{nmi} \frac{1}{k^{i+1}}e^{nkt} t^{m-i}
    & \text{if } n\neq 0, m \neq 0,\\
   \frac{1}{nk}e^{nkt}t^0 & \text{if } n\neq 0, m = 0,\\
   \frac{1}{m+1}e^{0kt}t^{m+1} & \text{if } n = 0, m \neq 0,\\
   e^{0kt}t^1 & \text{if } n =0 , m=0,
   \end{cases}

where :math:`c_{nm0} \triangleq \frac{1}{n}` and

.. math::

   c_{nmi} \triangleq \frac{(-1)^{i}}{n^{i+1}} \prod_{j=m-i+1}^{m} j,
   \quad 1\le i \le m.

The coefficient :math:`c_{nmi}` is implemented in function
:py:func:`~ajdmom.ito_mom.c_nmi` which returns a :class:`fractions.Fraction`
instead of a decimal (float number).

For the definite integral,

.. math::

   \int_{(n-1)h}^t e^{ik[s-(n-1)h]}[s-(n-1)h]^jds = F(t-(n-1)h) - F(0)

where :math:`F(t) = \int e^{nkt} t^m dt`. The definite integral is implemented
in :py:func:`~ajdmom.ito_mom.int_et`.


Polynomial Representation
--------------------------

The result of the integral, returned by :py:func:`~ajdmom.ito_mom.int_et`,
is represented as a "polynomial" of the following form

.. math::

   \int_{(n-1)h}^t e^{ik[s-(n-1)h]} [s-(n-1)h]^j ds
    = \sum_{i,j^{'},l}c_{ij^{'}l}e^{ik[t-(n-1)h]}[t-(n-1)h]^{j^{'}}k^{-l}

which is encoded in a :py:class:`~ajdmom.poly.Poly`, derived from
:class:`collections.UserDict`, with
:code:`keyfor = ('e^{k[t-(n-1)h]}','[t-(n-1)h]','k^{-}')`,
``key`` = :math:`(i,j^{'},l)` and ``value`` = :math:`c_{ij^{'}l}`.


Code Design
============

Itô process moment - I
-----------------------

With :math:`\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}|v_{n-1}]` represented as
a "polynomial" of the following form

.. math::

   &\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]\\
   &= \sum_{n_3,i,j,l,o,p,q} b_{n_3ijlopq} e^{n_3k(n-1)h} e^{ik[t-(n-1)h]}
   [t-(n-1)h]^jv_{n-1}^l k^{-o}\theta^p\sigma_v^q,

consequently, we have

.. math::

   &e^{-kt}\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]\\
   &= \sum_{n_3,i,j,l,o,p,q} b_{n_3ijlopq} e^{(n_3-1)k(n-1)h}
   e^{(i-1)k[t-(n-1)h]}[t-(n-1)h]^jv_{n-1}^l k^{-o}\theta^p\sigma_v^q,\\
   &e^{kt}\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]\\
   &= \sum_{n_3,i,j,l,o,p,q} b_{n_3ijlopq} e^{(n_3+1)k(n-1)h}
   e^{(i+1)k[t-(n-1)h]}[t-(n-1)h]^jv_{n-1}^l k^{-o}\theta^p\sigma_v^q,\\
   &e^{2kt}\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]\\
   &= \sum_{n_3,i,j,l,o,p,q} b_{n_3ijlopq} e^{(n_3+2)k(n-1)h}
   e^{(i+2)k[t-(n-1)h]}[t-(n-1)h]^jv_{n-1}^l k^{-o}\theta^p\sigma_v^q.

Therefore, it's profitable to consider the following generic integral

.. math::

   &\int_{(n-1)h}^t e^{mks}\mathbb{E}[I\!E_{n-1,s}^{n_3}I_{n-1,s}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]ds\\
   &= \sum_{n_3,i,j,l,o,p,q} b_{n_3ijlopq} e^{(n_3+m)k(n-1)h}
   \cdot int\_et(i+m,j)\cdot v_{n-1}^l k^{-o}\theta^p\sigma_v^q\\
   &= \sum_{n_3+m,i+m,j^{'},l,o^{'},p,q} b_{(n_3+m)(i+m)j^{'}l o^{'}pq}
   e^{(n_3+m)k(n-1)h} e^{(i+m)k[t-(n-1)h]} [t-(n-1)h]^{j^{'}}\\
   &\qquad \cdot v_{n-1}^{l} k^{-o^{'}}\theta^{p}\sigma_v^{q}

where

.. math::

   int\_et(i+m,j)
   =\sum_{i+m,j^{'},l^{'}} c_{(i+m)j^{'}l^{'}}e^{(i+m)k[t-(n-1)h]}
   [t-(n-1)h]^{j^{'}} k^{-l^{'}}.

Implementation:

1. Function :py:func:`~ajdmom.ito_mom.int_e_poly` in module
   :py:mod:`~ajdmom.ito_mom` is defined to accomplish the computation in
   equation :eq:`int-e-poly` of :doc:`../design`.

2. Function :py:func:`~ajdmom.ito_mom.recursive_IEII` in module
   :py:mod:`~ajdmom.ito_mom` is defined to realize
   the recursive step in equation :eq:`ito-moment` of :doc:`../theory`.

3. Function :py:func:`~ajdmom.ito_mom.moment_IEII` in module
   :py:mod:`~ajdmom.ito_mom` is implemented to calculate
   :math:`\mathbb{E}[I\!E_n^{n_3}I_n^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]`.

For demonstration, I re-write the following initial three moments in
:ref:`ito-recursive-i` in :doc:`../theory` according to the "polynomial"
representation

.. math::

   \mathbb{E}[I\!E_{n-1,t}^2|v_{n-1}]
   &=& \frac{1}{2}&e^{2k(n-1)h} e^{2k[t-(n-1)h]}[t-(n-1)h]^0v_{n-1}^0
   k^{-1}\theta^1\sigma_v^0\\
   && + &e^{2k(n-1)h}e^{k[t-(n-1)h]}[t-(n-1)h]^0v_{n-1}^1
   k^{-1}\theta^0\sigma_v^0\\
   && - &e^{2k(n-1)h}e^{k[t-(n-1)h]}[t-(n-1)h]^0v_{n-1}^0
   k^{-1}\theta^1\sigma_v^0\\
   && - &e^{2k(n-1)h}e^{0k[t-(n-1)h]}[t-(n-1)h]^0v_{n-1}^1
   k^{-1}\theta^0\sigma_v^0\\
   && + \frac{1}{2} &e^{2k(n-1)h}e^{0k[t-(n-1)h]}[t-(n-1)h]^0v_{n-1}^0
   k^{-1}\theta^1\sigma_v^0,\\
   %
   \mathbb{E}[I\!E_{n-1,t}I_{n-1,t}|v_{n-1}]
   &=& &e^{k(n-1)h} e^{k[t-(n-1)h]}[t-(n-1)h]^0v_{n-1}^0
   k^{-1}\theta^1\sigma_v^0\\
   && +&e^{k(n-1)h} e^{0k[t-(n-1)h]}[t-(n-1)h]^1v_{n-1}^1
   k^{-0}\theta^0\sigma_v^0\\
   && -&e^{k(n-1)h} e^{0k[t-(n-1)h]}[t-(n-1)h]^1v_{n-1}^0
   k^{-0}\theta^1\sigma_v^0\\
   && -&e^{k(n-1)h} e^{0k[t-(n-1)h]}[t-(n-1)h]^0v_{n-1}^0
   k^{-1}\theta^1\sigma_v^0,\\
   %
   \mathbb{E}[I_{n-1,t}^2|v_{n-1}]
   &=&-&e^{0k(n-1)h} e^{-k[t-(n-1)h]}[t-(n-1)h]^0v_{n-1}^1
   k^{-1}\theta^0\sigma_v^0\\
   && +&e^{0k(n-1)h} e^{-k[t-(n-1)h]}[t-(n-1)h]^0v_{n-1}^0
   k^{-1}\theta^1\sigma_v^0\\
   && +&e^{0k(n-1)h} e^{0k[t-(n-1)h]}[t-(n-1)h]^1v_{n-1}^0
   k^{-0}\theta^1\sigma_v^0\\
   && +&e^{0k(n-1)h} e^{0k[t-(n-1)h]}[t-(n-1)h]^0v_{n-1}^1
   k^{-1}\theta^0\sigma_v^0\\
   && -&e^{0k(n-1)h} e^{0k[t-(n-1)h]}[t-(n-1)h]^0v_{n-1}^0
   k^{-1}\theta^1\sigma_v^0.

"""
import math
from fractions import Fraction as Frac

from ajdmom.poly import Poly


def c_nmi(n, m, i):
    """Coefficent :math:`c_{nmi}` as in :eq:`c-nmi`.

    :param int n: n in :math:`e^{nkt}`.
    :param int m: m in :math:`t^{m-i}`.
    :param int i: i in :math:`t^{m-i}`.
    :return: the coefficient :math:`c_{nmi}`.
    :rtype: Fraction
    """
    prod = 1
    for j in range(m - i + 1, m + 1):
        prod = prod * j
    den = n ** (i + 1)  # denumerator
    c = ((-1) ** i) * Frac(prod, den)
    return c


def int_et(n, m):
    r""":math:`\int_{(n-1)h}^{t} e^{ik[s-(n-1)h]}[s-(n-1)h]^jds`

    :param int n: i in :math:`e^{ik[s-(n-1)h]}[s-(n-1)h]^j`.
    :param int m: j in :math:`e^{ik[s-(n-1)h]}[s-(n-1)h]^j`.
    :return: a poly with attribute ``keyfor`` =
       ('e^{k[t-(n-1)h]}', '[t-(n-1)h]', 'k^{-}').
    :rtype: Poly
    """
    if m < 0:
        msg = f"m in int_et(n,m) equals {m}, however it shouldn't be negative!"
        raise ValueError(msg)
    #
    poly = Poly()
    poly.set_keyfor(['e^{k[t-(n-1)h]}', '[t-(n-1)h]', 'k^{-}'])
    #
    if n == 0:
        poly[(0, m + 1, 0)] = Frac(1, m + 1)
    elif n != 0 and m == 0:
        poly[(n, 0, 1)] = Frac(1, n)
        poly[(0, 0, 1)] = -Frac(1, n)  # - F(0)
    else:
        poly[(n, m, 1)] = Frac(1, n)
        for i in range(1, m + 1):
            c = c_nmi(n, m, i)
            poly[(n, m - i, i + 1)] = c
            if i == m:  # - F(0): - c_nmi
                poly[(0, 0, i + 1)] = -c  # - c * 1/k^{i+1}
    return poly


def int_e_poly(c, tp, m, poly):
    r"""Integral of :math:`c \times tp \times \int_{(n-1)h}^t e^{mks} poly ds`

    :param int c: coefficient to multiply with
    :param int tp: type of the multiplication,

       +----+----------------------------+
       | tp | multiply with              |
       +====+============================+
       | 1  | :math:`e^{k(n-1)h}v_{n-1}` |
       +----+----------------------------+
       | 2  | :math:`-e^{k(n-1)h}\theta` |
       +----+----------------------------+
       | 3  | :math:`\theta`             |
       +----+----------------------------+
       | 4  | :math:`\sigma_v`           |
       +----+----------------------------+

    :param int m: power of :math:`e^{ks}` in the integrand
    :param Poly poly: a Poly object, such as
       :math:`\mathbb{E}[I\!E_t^{n_3} I_t^{n_4} I_t^{*n_5}]`, with attribute
       ``keyfor`` = ('e^{k(n-1)h}', 'e^{k[t-(n-1)h]}', '[t-(n-1)h]',
       'v_{n-1}','k^{-}', 'theta', 'sigma_v').
    :return: poly with the same ``keyfor`` as that in the input poly
    :rtype: Poly
    """
    poln = Poly()
    poln.set_keyfor(poly.keyfor)  # poly = IEII[(n3, n4, n5)]
    for k1 in poly:
        # \int e^{ik[s-(n-1)h]} [s-(n-1)h]^j ds
        #
        # e^{mks} -> e^{mk[s-(n-1)h]} (× e^{mk(n-1)h} compensated back later)
        poly_sub = int_et(m + k1[1], k1[2])
        # ('e^{k[t-(n-1)h]}', '[t-(n-1)h]', 'k^{-}')
        #
        for k2 in poly_sub:
            val = c * poly[k1] * poly_sub[k2]
            #
            # k1[0]+m: × e^{mk(n-1)h} to compensate e^{mk[s-(n-1)h]} in int_et(m + k1[1], k1[2])
            key = (k1[0] + m, k2[0], k2[1], k1[3], k1[4] + k2[2], k1[5], k1[6])
            #
            key = list(key)
            if tp == 1:
                key[0] += 1
                key[3] += 1
            elif tp == 2:
                key[0] += 1
                key[5] += 1
                val = -val
            elif tp == 3:
                key[5] += 1
            else:
                key[6] += 1
            key = tuple(key)
            poln.add_keyval(key, val)
    return poln


def recursive_IEII(n3, n4, n5, IEII):
    r"""Recursive step in equation :eq:`ito-moment`

    :param int n3: :math:`n_3`
       in :math:`\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]`.
    :param int n4: :math:`n_4`
       in :math:`\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]`.
    :param int n5: :math:`n_5`
       in :math:`\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]`.
    :param dict IEII: a dict with key (n3,n4,n5) and value Poly object with attribute
       ``keyfor`` =
       ('e^{k(n-1)h}', 'e^{k[t-(n-1)h]}', '[t-(n-1)h]', 'v_{n-1}',
       'k^{-}', 'theta', 'sigma_v').
    :return: updated IEII.
    :rtype: dict
    """
    kf = ['e^{k(n-1)h}', 'e^{k[t-(n-1)h]}', '[t-(n-1)h]',
          'v_{n-1}', 'k^{-}', 'theta', 'sigma_v']
    poly = Poly()
    poly.set_keyfor(kf)
    if n3 >= 2 and n4 >= 0 and n5 >= 0:
        c = Frac(n3 * (n3 - 1), 2)
        poly.merge(int_e_poly(c, 1, 1, IEII[(n3 - 2, n4, n5)]))
        poly.merge(int_e_poly(c, 2, 1, IEII[(n3 - 2, n4, n5)]))
        poly.merge(int_e_poly(c, 3, 2, IEII[(n3 - 2, n4, n5)]))
        poly.merge(int_e_poly(c, 4, 1, IEII[(n3 - 1, n4, n5)]))
    if n3 >= 0 and n4 >= 2 and n5 >= 0:
        c = Frac(n4 * (n4 - 1), 2)
        poly.merge(int_e_poly(c, 1, -1, IEII[(n3, n4 - 2, n5)]))
        poly.merge(int_e_poly(c, 2, -1, IEII[(n3, n4 - 2, n5)]))
        poly.merge(int_e_poly(c, 3, 0, IEII[(n3, n4 - 2, n5)]))
        poly.merge(int_e_poly(c, 4, -1, IEII[(n3 + 1, n4 - 2, n5)]))
    if n3 >= 1 and n4 >= 1 and n5 >= 0:
        c = Frac(n3 * n4, 1)
        poly.merge(int_e_poly(c, 1, 0, IEII[(n3 - 1, n4 - 1, n5)]))
        poly.merge(int_e_poly(c, 2, 0, IEII[(n3 - 1, n4 - 1, n5)]))
        poly.merge(int_e_poly(c, 3, 1, IEII[(n3 - 1, n4 - 1, n5)]))
        poly.merge(int_e_poly(c, 4, 0, IEII[(n3, n4 - 1, n5)]))
    if n3 >= 0 and n4 >= 0 and n5 >= 2:
        c = Frac(n5 * (n5 - 1), 2)
        poly.merge(int_e_poly(c, 1, -1, IEII[(n3, n4, n5 - 2,)]))
        poly.merge(int_e_poly(c, 2, -1, IEII[(n3, n4, n5 - 2)]))
        poly.merge(int_e_poly(c, 3, 0, IEII[(n3, n4, n5 - 2)]))
        poly.merge(int_e_poly(c, 4, -1, IEII[(n3 + 1, n4, n5 - 2)]))
    return poly


def moment_IEII(n3, n4, n5):
    r""":math:`\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]`

    :param int n3: :math:`n_3`
       in :math:`\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]`.
    :param int n4: :math:`n_4`
       in :math:`\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]`.
    :param int n5: :math:`n_5`
       in :math:`\mathbb{E}[I\!E_{n-1,t}^{n_3}I_{n-1,t}^{n_4}I_{n-1,t}^{*n_5}|v_{n-1}]`.
    :return: poly with attribute ``keyfor`` =
       ('e^{k(n-1)h}','e^{k[t-(n-1)h]}','[t-(n-1)h]','v_{n-1}','k^{-}','theta',
       'sigma_v').
    :rtype: Poly
    """
    if n3 + n4 + n5 < 0:
        raise ValueError(f"moment_IEII({n3},{n4},{n5}) is called!")
    #
    # IEII: a dict of moments of E[IE^{n3}I^{n4}I^{*n5}]
    #
    IEII = {}
    #
    kf = ['e^{k(n-1)h}', 'e^{k[t-(n-1)h]}', '[t-(n-1)h]', 'v_{n-1}', 'k^{-}',
          'theta', 'sigma_v']
    #
    # special poly constants, analog to 0 and 1
    #
    P0 = Poly({(0, 0, 0, 0, 0, 0, 0, 0): Frac(0, 1)})
    P0.set_keyfor(kf)
    P1 = Poly({(0, 0, 0, 0, 0, 0, 0, 0): Frac(1, 1)})
    P1.set_keyfor(kf)
    #
    # n3 + n4 + n5 = 0: special case
    #
    IEII[(0, 0, 0)] = P1  # equiv to constant 1
    #
    # n3 + n4 + n5 = 1
    #
    IEII[(1, 0, 0)] = P0  # equiv to constant 0
    IEII[(0, 1, 0)] = P0
    IEII[(0, 0, 1)] = P0
    #
    # n3 + n4 + n5 = 2
    #
    P200 = Poly({
        (2, 2, 0, 0, 1, 1, 0): Frac(1, 2),
        (2, 1, 0, 1, 1, 0, 0): Frac(1, 1),
        (2, 1, 0, 0, 1, 1, 0): Frac(-1, 1),
        (2, 0, 0, 1, 1, 0, 0): Frac(-1, 1),
        (2, 0, 0, 0, 1, 1, 0): Frac(1, 2)
    })
    P200.set_keyfor(kf)
    P110 = Poly({
        (1, 1, 0, 0, 1, 1, 0): Frac(1, 1),
        (1, 0, 1, 1, 0, 0, 0): Frac(1, 1),
        (1, 0, 1, 0, 0, 1, 0): Frac(-1, 1),
        (1, 0, 0, 0, 1, 1, 0): Frac(-1, 1)
    })
    P110.set_keyfor(kf)
    P020 = Poly({
        (0, -1, 0, 1, 1, 0, 0): Frac(-1, 1),
        (0, -1, 0, 0, 1, 1, 0): Frac(1, 1),
        (0, 0, 1, 0, 0, 1, 0): Frac(1, 1),
        (0, 0, 0, 1, 1, 0, 0): Frac(1, 1),
        (0, 0, 0, 0, 1, 1, 0): Frac(-1, 1)
    })
    P020.set_keyfor(kf)
    #
    IEII[(2, 0, 0)] = P200
    IEII[(1, 1, 0)] = P110
    IEII[(1, 0, 1)] = P0
    IEII[(0, 2, 0)] = P020
    IEII[(0, 1, 1)] = P0
    IEII[(0, 0, 2)] = P020
    #
    if n3 + n4 + n5 <= 2:
        return (IEII[(n3, n4, n5)])
    #
    # n1 + n2 + n3 >= 3: typical cases
    #
    if n3 + n4 + n5 > 3:
        # compute all lower-order moments to get ready
        for n in range(3, n3 + n4 + n5):
            for i in range(n, -1, -1):
                for j in range(n - i, -1, -1):
                    poly = recursive_IEII(i, j, n - i - j, IEII)
                    poly.remove_zero()
                    IEII[(i, j, n - i - j)] = poly
            # delete polys no more needed
            index = [key for key in IEII if key[0] + key[1] + key[2] == n - 2]
            for key in index: del IEII[key]
    # the last one
    poly = recursive_IEII(n3, n4, n5, IEII)
    poly.remove_zero()
    return poly


def moment_v(n):
    """Moment of :math:`v_{n-1}` as in equation :eq:`moment-v`

    :param int n: order of the moment.
    :return: a poly with attribute ``keyfor`` = ('theta','sigma_v^2/k').
    :rtype: Poly
    """
    v = []
    kf = ['theta', 'sigma_v^2/k']
    # n = 0
    poly = Poly({(0, 0): Frac(1, 1)})
    poly.set_keyfor(kf)
    v.append(poly)
    if n == 0: return (poly)
    # n = 1
    poly = Poly({(1, 0): Frac(1, 1)})
    poly.set_keyfor(kf)
    v.append(poly)
    if n == 1: return poly
    # n >= 2
    for i in range(2, n + 1):
        poly = v[-1]  # recursively computing
        poln = Poly()
        poln.set_keyfor(kf)
        for k in poly:
            # times theta and ((i-1)/2) sigma_v^2/k
            poln.add_keyval((k[0] + 1, k[1]), poly[k])
            poln.add_keyval((k[0], k[1] + 1), poly[k] * Frac(i - 1, 2))
        v.append(poln)
    return poln


def poly2num(poly, par):
    """Convert a polynomial to numerical value

    :param Poly poly: polynomial to convert, with attribute ``keyfor`` =
       ('e^{k(n-1)h}','e^{k[t-(n-1)h]}','[t-(n-1)h]','v_{n-1}','k^{-}','theta',
       'sigma_v').
    :param dict par: parameters.
    :return: numerical value.
    :rtype: float
    """
    k, h, v0, theta, sigma = par['k'], par['h'], par['v0'], par['theta'], par['sigma']
    f = 0
    for K in poly:
        # (n-1) = 0, [t-(n-1)h] = h
        val = math.exp(K[1] * k * h) * (h ** K[2]) * (v0 ** K[3]) * (k ** (-K[4]))
        val *= (theta ** K[5]) * (sigma ** K[6])
        f += val * poly[K]
    return f


if __name__ == "__main__":
    # Example usage of the module, see 'tests/test_ito_mom.py' for more test
    from pprint import pprint
    import sys

    print('\nExample usage of the module functions\n')
    #
    kf = ('theta', 'sigma_v^2/k')
    print(f'\nmoment_v(n) returns a poly with keyfor = \n{kf}')
    print(f'moment_v(4) = ')
    pprint(moment_v(4))
    #
    kf = ('e^{k(n-1)h}', 'e^{k[t-(n-1)h]}', '[t-(n-1)h]', 'v_{n-1}', 'k^{-}',
          'theta', 'sigma_v')
    print(f"\nmoment_IEII(n3,n4,n5) returns a poly with keyfor = \n{kf}")
    args = sys.argv[1:]
    n = 3 if len(args) == 0 else int(args[0])
    #
    for n3 in range(n, -1, -1):
        for n4 in range(n - n3, -1, -1):
            n5 = n - n3 - n4
            print(f'\nmoment_IEII({n3},{n4},{n5}) = ')
            pprint(moment_IEII(n3, n4, n5))



import numpy as np
import pytest

from ajdmom.mdl_1fsvj.euler import r1FSVJ
from ajdmom.mdl_1fsvj.mom import m
from ajdmom.mdl_1fsvj.cmom import cm
from ajdmom.mdl_1fsvj.cov import cov


@pytest.fixture
def settings():
    par = {'mu': 0.125, 'k': 0.1, 'theta': 0.25, 'sigma_v': 0.1, 'rho': -0.7,
           'lambda': 0.01, 'mu_j': 0, 'sigma_j': 0.5, 'h': 1}
    v0 = par['theta']
    y_series = r1FSVJ(v0, par, N=4000 * 1000, n_segment=10, h=1)
    # threshold for reporting discrepancy
    rel_err = 0.100  # 10%
    abs_err = 0.001
    return par, y_series, rel_err, abs_err


def test_m(settings):
    """Test moment computing"""
    par, y_series, rel_err, abs_err = settings
    #
    ns = [1, 2, 3, 4, 5]
    print("")
    for n in ns:
        expected = np.mean(y_series ** n)
        actual = m(n, par)
        expected = pytest.approx(expected, rel=rel_err, abs=abs_err)
        msg = "Diff in the {}th moment is > {:} or {:.0%} "
        msg += "between theory and sample."
        devi = max(expected.abs, expected.rel * abs(expected.expected))
        temp = "{:d}-th moment: actual({: f}) V.S. expected({: f} ± {:f})"
        print(temp.format(n, actual, expected.expected, devi))
        assert actual == expected, msg.format(n, abs_err, rel_err)


def test_cm(settings):
    """Test central moment computing"""
    par, y_series, rel_err, abs_err = settings
    #
    ns = [2, 3, 4, 5]
    diff = y_series - np.mean(y_series)
    del y_series
    print("")
    for n in ns:
        expected = np.mean(diff ** n)
        actual = cm(n, par)
        expected = pytest.approx(expected, rel=rel_err, abs=abs_err)
        msg = "Diff in the {}th central moment is > {:} or {:.0%} "
        msg += "between theory and sample."
        devi = max(expected.abs, expected.rel * abs(expected.expected))
        temp = "{:d}-th central moment: actual({: f}) V.S. expected({: f} ± {:f})"
        print(temp.format(n, actual, expected.expected, devi))
        assert actual == expected, msg.format(n, abs_err, rel_err)


def test_cov(settings):
    """Test covariance computing"""
    par, y_series, rel_err, abs_err = settings
    #
    N = len(y_series)
    pre = y_series[0:(N - 1)]
    nxt = y_series[1:N]
    #
    nms = [(1, 1), (2, 1), (1, 2), (3, 1), (2, 2), (1, 3), (4, 1), (3, 2), (2, 3), (1, 4)]
    print("")
    for nm in nms:
        n, m = nm
        expected = np.cov(pre ** n, nxt ** m)[0][1]
        actual = cov(n, m, par)
        expected = pytest.approx(expected, rel=rel_err, abs=abs_err)
        msg = "Diff in the lag-1 autocov({},{}) is > {:} or {:.0%} "
        msg += "between theory and sample."
        devi = max(expected.abs, expected.rel * abs(expected.expected))
        temp = "cov({:d},{:d}): actual({: f}) V.S. expected({: f} ± {:f})"
        print(temp.format(n, m, actual, expected.expected, devi))
        assert actual == expected, msg.format(n, m, abs_err, rel_err)



import pytest
from ajdmom.poly import Poly


@pytest.fixture
def two_polys():
    pol1 = Poly({(1, 0): 1, (0, 1): 1})
    pol2 = Poly({(1, 0): 1, (0, 2): 1})
    pol1.set_keyfor(['e', 'h'])
    pol2.set_keyfor(['e', 'h'])
    #
    return ((pol1, pol2))


def test_add(two_polys):
    """Test Poly addition operation"""
    expected = Poly({(1, 0): 2, (0, 1): 1, (0, 2): 1})
    expected.set_keyfor(['e', 'h'])
    #
    pol1, pol2 = two_polys
    actual = pol1 + pol2
    #
    assert actual == expected, "Poly addition operation is wrong!"
    #
    msg = "commutative law of addition is broken!"
    assert pol1 + pol2 == pol2 + pol1, msg


def test_merge(two_polys):
    """Test Poly merge operation"""
    expected = Poly({(1, 0): 2, (0, 1): 1, (0, 2): 1})
    expected.set_keyfor(['e', 'h'])
    #
    pol1, pol2 = two_polys
    pol1.merge(pol2)
    actual = pol1
    #
    assert actual == expected, "Poly merge operation is wrong!"


def test_sub(two_polys):
    """Test Poly subtraction operation"""
    expected = Poly({(0, 1): 1, (0, 2): -1})
    expected.set_keyfor(['e', 'h'])
    #
    pol1, pol2 = two_polys
    actual = pol1 - pol2
    #
    assert actual == expected, "Poly subtraction operation is wrong!"


def test_mul(two_polys):
    """Test Poly multiplication operation"""
    expected = Poly({(2, 0): 1, (1, 2): 1, (1, 1): 1, (0, 3): 1})
    expected.set_keyfor(['e', 'h'])
    #
    pol1, pol2 = two_polys
    actual = pol1 * pol2
    #
    assert actual == expected, "Poly multiplication operation is wrong!"


def test_rmul(two_polys):
    """Test Poly reverse multiplication operation"""
    expected = Poly({(1, 0): 4, (0, 1): 4})
    expected.set_keyfor(['e', 'h'])
    #
    pol1, pol2 = two_polys
    actual = 4 * pol1
    #
    assert actual == expected, "Poly reverse multiplication is wrong!"


def test_pow(two_polys):
    """Test Poly power operation"""
    expected = Poly({(2, 0): 1, (1, 1): 2, (0, 2): 1})
    expected.set_keyfor(['e', 'h'])
    #
    pol1, pol2 = two_polys
    actual = pol1 ** 2
    #
    assert actual == expected, "Poly power is wrong!"
    #
    expected = Poly({(3, 0): 1, (2, 1): 3, (1, 2): 3, (0, 3): 1})
    expected.set_keyfor(['e', 'h'])
    #
    actual = pol1 ** 3
    #
    assert actual == expected, "Poly power is wrong still!"


def test_mul_poly(two_polys):
    """Test Poly multiplication with a different poly"""
    poln = Poly({(1,): 100, (2, 0): 100})
    #
    pol1, pol2 = two_polys
    #
    expected = Poly({(2, 0): 100, (1, 1): 100, (3, 0): 100, (2, 1): 100})
    #
    actual = pol1.mul_poly(poln, keyIndexes=([0, 1], [0, -1]), keyfor=('e', 'h'))
    #
    msg = "Poly multiplication of different type Poly is wrong!"
    assert actual == expected, msg



"""
Module for generating samples from :abbr:`SVCJ(Stochastic Volatility with
Contemporaneous Jumps in the price and variance)`

Note that we are generating i.i.d. samples with fixed conditions instead of
a trajectory of samples!
"""
import math

import numpy as np
from ajdmom.mdl_svvj.euler import rSVVJ

rng = np.random.default_rng()


def one_step_euler(v0, mu, k, theta, sigma, rho, tau):
    eps1 = rng.normal(0, 1)
    eps2 = rng.normal(0, 1)
    diffusion1 = sigma * math.sqrt(v0 * tau) * eps1
    diffusion2 = math.sqrt(v0 * tau) * (rho * eps1 + math.sqrt(1-rho**2) * eps2)
    #
    v = v0 + k * (theta - v0) * tau + diffusion1
    v = v if v >= 0 else 0
    #
    y = (mu - v0/2) * tau + diffusion2
    return v, y

def mul_step_euler(v0, mu, k, theta, sigma, rho, tau, nsegment=10):
    delta = tau / nsegment
    cumy = 0
    if nsegment == 0:
        v, y = one_step_euler(v0, mu, k, theta, sigma, rho, tau)
        return v, y
    for i in range(nsegment):
        v, y = one_step_euler(v0, mu, k, theta, sigma, rho, delta)
        v0 = v
        cumy += y
    return v, cumy

def r_v_y(v0, mu, k, theta, sigma, rho, lmbd, mu_v, rhoJ, mu_s, sigma_s, h, nsegment=10):
    numJ = rng.poisson(lmbd * h)
    jtime = sorted(rng.uniform(0, h, numJ))
    jtime = np.array(jtime)
    jsizev = rng.exponential(mu_v, numJ)
    jsizey = rng.normal(mu_s, sigma_s, numJ) + rhoJ * jsizev
    #
    cumy = 0
    #
    if numJ == 0:
        v, cumy = mul_step_euler(v0, mu, k, theta, sigma, rho, h, nsegment)
        return v, cumy
    # fine simulation
    for i in range(numJ):
        delta = jtime[0] if i == 0 else jtime[i] - jtime[i-1]
        ratio = delta / h
        # diffusion
        nseg = math.ceil(ratio * nsegment)
        v, y = mul_step_euler(v0, mu, k, theta, sigma, rho, delta, nseg)
        # jump
        v += jsizev[i]
        y += jsizey[i]
        cumy += y
        #
        v0 = v
    # last diffusion
    delta = h - jtime[-1]
    ratio = delta / h
    nseg = math.ceil(ratio * nsegment)
    v, y = mul_step_euler(v0, mu, k, theta, sigma, rho, delta, nseg)
    cumy += y
    return v, cumy

def r_y(v0, mu, k, theta, sigma, rho, lmbd, mu_v, rhoJ, mu_s, sigma_s, h, N, nsegment=10):
    """Euler approximation"""
    # y_series = []
    ys = []
    for i in range(N):
        # v0, y = r_v_y(v0, mu, k, theta, sigma, rho, lmbd, mu_v, rhoJ, mu_s, sigma_s, h, nsegment)
        # y_series.append(y)
        v, y = r_v_y(v0, mu, k, theta, sigma, rho, lmbd, mu_v, rhoJ, mu_s, sigma_s, h, nsegment)
        ys.append(y)
    # return y_series
    return ys

def rSVCJ(v0, mu, k, theta, sigma, rho, T, jsize_v, jtime_v,
          mu_s, sigma_s, rhoJ, n, n_segment):
    r"""Generate i.i.d. Random samples from mdl_svcj by Euler approximation

    :param float v0: value of the initial variance :math:`v(0)`
    :param float mu: parameter :math:`\mu`
    :param float k: parameter :math:`k`
    :param float theta: parameter :math:`\theta`
    :param float sigma: parameter :math:`\sigma_v`
    :param float rho: parameter :math:`\rho`
    :param float T: end time
    :param tuple jsize_v: jump sizes in the variance over [0, T]
    :param tuple jtime_v: jump time points in the variance over [0, T]
    :param float mu_s: jumps in the price distributed according to
      :math:`\mathcal{N}(\mu_s + J^v\rhoJ, \sigma_s^2)`
    :param float sigma_s: standard deviation of the normal distribution of jumps
      in the price
    :param float rhoJ: :math:`\mu_s + J^v\rhoJ`
    :param int n: number of samples to generate
    :param int n_seqment: number of segments to split [0, T]
    :return: i.i.d. samples
    :rtype: numpy.ndarray
    """
    y_svvj = rSVVJ(v0, mu, k, theta, sigma, rho, T, jsize_v, jtime_v, n, n_segment)
    #
    IZ = np.zeros(n)
    for i in range(n):
        for j in range(len(jsize_v)):
            IZ[i] += rng.normal(mu_s + rhoJ * jsize_v[j], sigma_s)
    return y_svvj + IZ


if __name__ == '__main__':
    import sys

    v0, h = 0.007569, 1
    k, theta, sigma, rho = 3.46, 0.008, 0.14, -0.82
    r = 0.0319
    lmbd_v, mu_v = 0.47, 0.05
    mu_b = -0.1
    lmbd_s, sigma_s, rhoJ = lmbd_v, 0.0001, -0.38
    #
    mu_s = math.log((1 + mu_b) * (1 - rhoJ * mu_v)) - sigma_s ** 2 / 2
    #
    args = sys.argv[1:]
    numJ = rng.poisson(lmbd_v * h) if len(args) == 0 else int(args[0])
    jsize_v = rng.exponential(mu_v, numJ)
    jtime_v = rng.uniform(0, h, numJ)
    jtime_v = sorted(jtime_v)
    # change to tuples
    jsize_v, jtime_v = tuple(jsize_v), tuple(jtime_v)
    #
    n, n_segment = 20, 10
    y = rSVCJ(v0, r, k, theta, sigma, rho, h, jsize_v, jtime_v,
              mu_s, sigma_s, rhoJ, n, n_segment)
    print(f"\nGiven v0 = {v0} and {numJ} jumps in the variance, ")
    print(f"jsize_v = {jsize_v}, \njtime_v = {jtime_v}")
    print(f"\n{n} i.i.d. samples from SVCJ: ")
    print(y)



"""
Conditional central moments (II)

Conditional central moments of the SRJD model, given the initial state
and the realized jumps over the concerned interval.
"""
from ajdmom.poly import Poly
from ajdmom.mdl_srjd.cond2_mom import moment_IE, poly2num


def cmoment_v(m):
    r"""conditional central moment :math:`\mathbb{E}[\overline{v}^m(t)
    |v_0, z(s), 0\le s\le t]`

    :param int m: order of the conditional central moments of :math:`v(t)`
    :return: poly with attribute ``keyfor`` =
      ('v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}')
    :rtype: Poly
    """
    poly = Poly()
    kf = ('v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}')
    poly.set_keyfor(kf)
    #
    poln = moment_IE(m)
    # ('v_0', 'k^{-}', 'theta', 'sigma', 'e^{kt}', 'l_{1:n}')
    for k in poln:
        key = list(k)
        key[3] += m
        key[4] -= m
        key = tuple(key)
        val = poln[k]
        poly.add_keyval(key, val)
    return poly


##########
# scalar
##########
def cm(l, par):
    """conditional central moment in scalar

    :param int l: order of the conditional central moment
    :param dict par: parameters in dict, ``jumptime`` and ``jumpsize`` must also be
      included
    :return: scalar of the central moment
    :rtype: float
    """
    cmoment = cmoment_v(l)
    value = poly2num(cmoment, par)
    return value


if __name__ == "__main__":
    from pprint import pprint
    import sys

    print('\nExample usage of the module function\n')
    args = sys.argv[1:]
    n = 2 if len(args) == 0 else int(args[0])
    cmom = cmoment_v(n)
    print(f"cmoment_v({n}) = \n")
    pprint(cmom)
    print(f"\nwhich is a Poly with attribute keyfor = \n{cmom.keyfor}")



import numpy as np
import pytest
import math

from ajdmom.mdl_svij.euler import rSVIJ
from ajdmom.mdl_svij.cond2_cmom import cm
from ajdmom.mdl_svij.cond2_mom import m

rng = np.random.default_rng()


@pytest.fixture
def settings():
    v0, h = 0.007569, 1
    mu, k, theta, sigma, rho = 0.0319, 3.46, 0.008, 0.14, -0.82
    lmbd_v, mu_v = 0.47, 0.05  # same as that in mdl_svcj
    #
    mu_b = -0.12
    lmbd_s, sigma_s = 0.11, 0.15
    mu_s = math.log(1 + mu_b) - sigma_s ** 2 / 2
    # same as that in SVJ in Broadie-Kaya (2006)
    rel_err = 0.100 # 10%
    abs_err = 0.001
    par = {'v0': v0, 'h': h, 'mu': mu, 'k': k, 'theta': theta,
           'sigma': sigma, 'rho': rho, 'lmbd_v': lmbd_v, 'mu_v': mu_v,
           'lmbd_s': lmbd_s, 'mu_s': mu_s, 'sigma_s': sigma_s}
    return par, rel_err, abs_err


def test_rSVIJ(settings):
    par, rel_err, abs_err = settings
    #
    k, theta, sigma, rho = par['k'], par['theta'], par['sigma'], par['rho']
    v0, h, mu = par['v0'], par['h'], par['mu']
    lmbd_v, mu_v = par['lmbd_v'], par['mu_v']
    lmbd_s, mu_s, sigma_s = par['lmbd_s'], par['mu_s'], par['sigma_s']
    #
    for numJ in [0, 1, 2, 3]:
        jsize_v = rng.exponential(mu_v, numJ)
        jtime_v = rng.uniform(0, h, numJ)  # unsorted
        jsize_v, jtime_v = tuple(jsize_v), tuple(sorted(jtime_v))  # set as tuples
        par['jumpsize'] = jsize_v
        par['jumptime'] = jtime_v
        n = 100 * 1000
        n_segment = [10, 50, 100]
        # n_segment = [1, 5, 10]
        # n_segment = range(1, 11)
        tm = [m(j, par) for j in [1, 2, 3, 4, 5]]
        print(f"\nnumJ = {numJ}, \njumpsize = {jsize_v}, \njumptime = {jtime_v}")
        temp = "tmean = [{: f}, {: f}, {: f}, {: f}, {: f}]"
        print(temp.format(tm[0], tm[1], tm[2], tm[3], tm[4]))
        for i in range(len(n_segment)):
            y = rSVIJ(v0, mu, k, theta, sigma, rho, h, jsize_v, jtime_v,
                      lmbd_s, mu_s, sigma_s, n, n_segment[i])
            sm = []
            for j in [1, 2, 3, 4, 5]:
                sm.append(np.mean(y ** j))
            temp = "smean = [{: f}, {: f}, {: f}, {: f}, {: f}], n_segment = {:d}"
            print(temp.format(sm[0], sm[1], sm[2], sm[3], sm[4], n_segment[i]))


def test_m(settings):
    par, rel_err, abs_err = settings
    #
    k, theta, sigma, rho = par['k'], par['theta'], par['sigma'], par['rho']
    v0, h, mu = par['v0'], par['h'], par['mu']
    lmbd_v, mu_v = par['lmbd_v'], par['mu_v']
    lmbd_s, mu_s, sigma_s = par['lmbd_s'], par['mu_s'], par['sigma_s']
    #
    for numJ in [0, 1, 2, 3]:  # numJ = rng.poisson(lmbd * h): the actual number of jumps
        jsize_v = rng.exponential(mu_v, numJ)
        jtime_v = rng.uniform(0, h, numJ)  # unsorted
        jsize_v, jtime_v = tuple(jsize_v), tuple(sorted(jtime_v))  # set as tuples
        par['jumpsize'] = jsize_v
        par['jumptime'] = jtime_v
        #
        n, n_segment = 100 * 1000, 100
        y = rSVIJ(v0, mu, k, theta, sigma, rho, h, jsize_v, jtime_v,
                  lmbd_s, mu_s, sigma_s, n, n_segment)
        print(f"\nWith conditions v0 = {v0}, \njumpsize = {jsize_v}, \njumptime = {jtime_v}")
        for n in [1, 2, 3, 4, 5]:
            expected = np.mean(y ** n)
            expected = pytest.approx(expected, rel=rel_err, abs=abs_err)
            actual = m(n, par)
            msg = "Diff in the {}th conditional moment is > {:} or {:.0%} "
            msg += "between theory and sample."
            devi = max(expected.abs, expected.rel * abs(expected.expected))
            temp = "{:d}-th moment: actual({: f}) V.S. expected({: f} ± {:f})"
            print(temp.format(n, actual, expected.expected, devi))
            assert actual == expected, msg.format(n, abs_err, rel_err)


def test_cm(settings):
    par, rel_err, abs_err = settings
    #
    k, theta, sigma, rho = par['k'], par['theta'], par['sigma'], par['rho']
    v0, h, mu = par['v0'], par['h'], par['mu']
    lmbd_v, mu_v = par['lmbd_v'], par['mu_v']
    lmbd_s, mu_s, sigma_s = par['lmbd_s'], par['mu_s'], par['sigma_s']
    #
    for numJ in [0, 1, 2, 3]:  # numJ = rng.poisson(lmbd * h): the actual number of jumps
        jsize_v = rng.exponential(mu_v, numJ)
        jtime_v = rng.uniform(0, h, numJ)  # unsorted
        jsize_v, jtime_v = tuple(jsize_v), tuple(sorted(jtime_v))  # set as tuples
        par['jumpsize'] = jsize_v
        par['jumptime'] = jtime_v
        #
        n, n_segment = 100 * 1000, 100
        y = rSVIJ(v0, mu, k, theta, sigma, rho, h, jsize_v, jtime_v,
                  lmbd_s, mu_s, sigma_s, n, n_segment)
        print(f"\nWith conditions v0 = {v0}, \njumpsize = {jsize_v}, \njumptime = {jtime_v}")
        diff = y - np.mean(y)
        del y
        for n in [2, 3, 4, 5]:
            expected = np.mean(diff ** n)
            expected = pytest.approx(expected, rel=rel_err, abs=abs_err)
            actual = cm(n, par)
            msg = "Diff in the {}th conditional central moment is > {:} or {:.0%} "
            msg += "between theory and sample."
            devi = max(expected.abs, expected.rel * abs(expected.expected))
            temp = "{:d}-th central moment: actual({: f}) V.S. expected({: f} ± {:f})"
            print(temp.format(n, actual, expected.expected, devi))
            assert actual == expected, msg.format(n, abs_err, rel_err)



import numpy as np
import pytest

from ajdmom.mdl_srjd.euler import rSRJD
from ajdmom.mdl_srjd.cond2_mom import m
from ajdmom.mdl_srjd.cond2_cmom import cm

rng = np.random.default_rng()


@pytest.fixture
def settings():
    v0, h = 0.007569, 1
    k, theta, sigma = 3.46, 0.008, 0.14
    mu_v, lmbd = 0.05, 0.47
    #
    rel_err = 0.100  # 10%
    abs_err = 0.001
    par = {'v0': v0, 'h': h, 'k': k, 'theta': theta, 'sigma': sigma, 'mu_v': mu_v, 'lmbd': lmbd}
    return par, rel_err, abs_err


def test_m(settings):
    par, rel_err, abs_err = settings
    #
    k, theta, sigma = par['k'], par['theta'], par['sigma']
    lmbd, mu_v = par['lmbd'], par['mu_v']
    v0, h = par['v0'], par['h']
    #
    for numJ in [0, 1, 2, 3]:  # numJ = rng.poisson(lmbd * h): the actual number of jumps
        jumpsize = rng.exponential(mu_v, numJ)
        jumptime = rng.uniform(0, h, numJ)  # unsorted
        jumpsize, jumptime = tuple(jumpsize), tuple(sorted(jumptime))  # set as tuples
        par['jumpsize'] = jumpsize
        par['jumptime'] = jumptime
        #
        x = rSRJD(v0, k, theta, sigma, h, jumpsize, jumptime, 100 * 1000)
        print(f"\nWith conditions v0 = {v0}, \njumpsize = {jumpsize}, \njumptime = {jumptime}")
        for n in [1, 2, 3, 4, 5]:
            expected = np.mean(x ** n)
            expected = pytest.approx(expected, rel=rel_err, abs=abs_err)
            actual = m(n, par)
            msg = "Diff in the {}th conditional moment is > {:} or {:.0%} "
            msg += "between theory and sample."
            devi = max(expected.abs, expected.rel * abs(expected.expected))
            temp = "{:d}-th moment: actual({: f}) V.S. expected({: f} ± {:f})"
            print(temp.format(n, actual, expected.expected, devi))
            assert actual == expected, msg.format(n, abs_err, rel_err)


def test_cm(settings):
    par, rel_err, abs_err = settings
    #
    k, theta, sigma = par['k'], par['theta'], par['sigma']
    lmbd, mu_v = par['lmbd'], par['mu_v']
    v0, h = par['v0'], par['h']
    #
    for numJ in [0, 1, 2, 3]:  # numJ = rng.poisson(lmbd * h): the actual number of jumps
        jumpsize = rng.exponential(mu_v, numJ)
        jumptime = rng.uniform(0, h, numJ)
        jumpsize, jumptime = tuple(jumpsize), tuple(sorted(jumptime))
        par['jumpsize'] = jumpsize
        par['jumptime'] = jumptime
        #
        x = rSRJD(v0, k, theta, sigma, h, jumpsize, jumptime, 100 * 1000)
        print(f"\nWith conditions v0 = {v0}, \njumpsize = {jumpsize}, \njumptime = {jumptime}")
        diff = x - np.mean(x)
        del x
        for n in [2, 3, 4, 5]:
            expected = np.mean(diff ** n)
            expected = pytest.approx(expected, rel=rel_err, abs=abs_err)
            actual = cm(n, par)
            msg = "Diff in the {}th conditional central moment is > {:} or {:.0%}"
            msg += "between theory and sample."
            devi = max(expected.abs, expected.rel * abs(expected.expected))
            temp = "{:d}-th central moment: actual({: f}) V.S. expected({: f} ± {:f})"
            print(temp.format(n, actual, expected.expected, devi))
            assert actual == expected, msg.format(n, abs_err, rel_err)



"""
Moments of :math:`IEZ_{s,t}`
"""
import math
from fractions import Fraction

from ajdmom.poly import Poly
from ajdmom.cpp_mom import dmgf_cpp

def moment_EJ(n):
    # kf = ['e^{kt}-e^{ks}', 't-s', 'k^{-}', 'mu_v']
    kf = ['e^{kt}', 'e^{ks}', 't-s', 'k^{-}', 'mu_v']
    poly = Poly()
    poly.set_keyfor(kf)
    if n < 0:
        raise ValueError(f'moment_EJ({n}) is called!')
    if n == 0:
        key = (0, 0, 0, 0, 0)
        val = Fraction(1, 1)
        poly.add_keyval(key, val)
        return poly
    key = (n, 0, -1, 1, n)
    val = Fraction(math.factorial(n-1), 1)
    poly.add_keyval(key, val)
    #
    key = (0, n, -1, 1, n)
    val = Fraction(-math.factorial(n-1), 1)
    poly.add_keyval(key, val)
    return poly

def decode(poly):
    poln = Poly()
    poln.set_keyfor(['e^{kt}', 'e^{ks}', 't-s', 'k^{-}', 'lmbd', 'mu_v'])
    for k in poly:
        i = k[0]
        #
        pol1 = Poly({(0, 0, 0, 0, 0): Fraction(1, 1)})
        pol1.set_keyfor(['e^{kt}', 'e^{ks}', 't-s', 'k^{-}', 'mu_v'])
        for j in range(1, len(k)):
            n_j, m_j = k[j]
            pol1 = pol1 * (moment_EJ(n_j) ** m_j)
        #
        for key in pol1:
            knw = (key[0], key[1], key[2] + i, key[3], i, key[4])
            val = poly[k] * pol1[key]
            poln.add_keyval(knw, val)
    poln.remove_zero()
    #
    pol2 = Poly()
    pol2.set_keyfor(['e^{kt}', 'e^{ks}', 'k^{-}', 'lmbd', 'mu_v'])
    for key in poln:
        if key[2] != 0:
            raise ValueError("power of 't-s' is not 0 in moment_IEZ(n)!")
    for k, v in poln.items():
        key = k[0:2] + k[3:]
        pol2.add_keyval(key, v)
    return pol2

def moment_IEZ(n):
    """Moment of :math:`E[IEZ_{s,t}^n]`

    :param int n: order of the moment
    :return: poly with ``keyfor`` = ('e^{kt}', 'e^{ks}', 'k^{-}', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    if n < 0:
        raise ValueError(f"moment_IEZ({n}) is called!")
    #
    M = []
    kf = ['[lmbd(t-s)]^{i} M^{(n_1)m_1}(s) ... M^{(n_l)m_l}(s)']
    # n = 0
    # kf = ['e^{kt}-e^{ks}', 't-s', 'k^{-}', 'lmbd', 'mu_v']
    poly = Poly({(0,): Fraction(1, 1)})
    poly.set_keyfor(kf)
    M.append(poly)
    #
    if n == 0:
        poln = Poly({(0, 0, 0, 0, 0): Fraction(1, 1)})
        poln.set_keyfor(['e^{kt}', 'e^{ks}', 'k^{-}', 'lmbd', 'mu_v'])
        return poln
    # n = 1
    poly = Poly({(1, (1, 1)): 1})
    poly.set_keyfor(kf)
    M.append(poly)
    #
    if n == 1:
        poln = Poly({
            (1, 0, 1, 1, 1): Fraction(1, 1),
            (0, 1, 1, 1, 1): Fraction(-1, 1)
        })
        poln.set_keyfor(['e^{kt}', 'e^{ks}', 'k^{-}', 'lmbd', 'mu_v'])
        return poln
    #
    for i in range(2, n+1):
        M.append(dmgf_cpp(M[-1]))
    return decode(M[-1])

def poly2num(poly, par):
    # ['e^{kt}', 'e^{ks}', 'k^{-}', 'lmbd', 'mu_v']
    k, lmbd, mu_v = par['k'], par['lmbd'], par['mu_v']
    s, t = par['lb'], par['ub']
    f = 0
    for K, v in poly.items():
        val = math.exp(K[0] * k * t) * math.exp(K[1] * k * s)
        val *= (k ** (-K[2])) * (lmbd ** K[3]) * (mu_v ** K[4])
        f += val * v
    return f

def m(n, par):
    moment = moment_IEZ(n)
    value = poly2num(moment, par)
    return value


if __name__ == '__main__':
    from pprint import pprint

    print('\nExample usage of the module functions')

    n = 3
    print(f"moment_EJ({n}) = ")
    pprint(moment_EJ(n))
    print("which is a poly with keyfor = ('e^{kt}', 'e^{ks}', 't-s', 'k^{-}', 'mu_v')\n")
    #
    # n = 1
    print(f"moment_IEZ({n}) = ")
    pprint(moment_IEZ(n))
    print("which is a poly with keyfor = ('e^{kt}', 'e^{ks}', 'k^{-}', 'lmbd', 'mu_v')")



r"""
Moments of the SVCJ model
"""
import math
from fractions import Fraction

from ajdmom.poly import Poly
from ajdmom.mdl_svcj.cond_mom import moment_y as cond_moment_y
from ajdmom.mdl_srjd.cmom import cmoment_v

def moment_v_theta(n):
    """moment of :math:`E[(v0-theta)^n]`

    :param int n: order of the moment
    :return: poly with ``keyfor`` = ('k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    kf = ['k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    #
    # v0 - theta = (v0 - E[v]) + lmbd * mu_v / k
    # E[(v0 - theta)^n] = \sum_{i=0}^n \binom{n}{i} (lmbd * mu_v / k)^i E[(v0 - E[v])^{n-i}]
    #
    poly = Poly()
    poly.set_keyfor(kf)
    for i in range(n + 1):
        bino = math.comb(n, i)
        uncmom = cmoment_v(n - i)
        # 'k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v'
        for k, v in uncmom.items():
            key = list(k)
            key[0] += i
            key[3] += i
            key[4] += i
            key = tuple(key)
            poly.add_keyval(key, v * bino)
    return poly

def key_times_poly(coef, k, poly):
    # k: ['e^{kt}', 't', 'k^{-}',
    #     'mu', 'v0-theta', 'theta', 'sigma', 'rho',
    #            substituted
    #     'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s']
    #
    # poly: ['k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    #
    kf = ['e^{kt}', 't', 'k^{-}',
          'mu', 'E[v]', 'theta', 'sigma', 'rho',
          'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s']
    #
    poln = Poly()
    poln.set_keyfor(kf)
    for K, V in poly.items():
        key = (k[0],         # e^{kt}
               k[1],         # t
               k[2] + K[0],  # k^{-}
               k[3],         # mu
                      K[1],  # E[v]
               k[5],         # theta
               k[6] + K[2],  # sigma
               k[7],         # rho
               k[8] + K[3],  # lmbd
               k[9] + K[4],  # mu_v
               k[10],        # rhoJ
               k[11],        # mu_s
               k[12])        # sigma_s
        poln.add_keyval(key, coef * V)
    return poln

def moment_y(n):
    """moment of :math:`y_t`

    :param int n: order of the moment
    :return: poly with ``keyfor`` = ('e^{kt}', 't', 'k^{-}',
      'mu', 'E[v]', 'theta', 'sigma', 'rho',
      'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s')
    :rtype: Poly
    """
    poly = cond_moment_y(n)
    # kf = ['e^{kt}', 't', 'k^{-}',
    #       'mu', 'v0-theta', 'theta', 'sigma', 'rho',
    #       'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s']
    kf = ['e^{kt}', 't', 'k^{-}',
          'mu', 'E[v]', 'theta', 'sigma', 'rho',
          'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s']
    poln = Poly()
    poln.set_keyfor(kf)
    #
    for k, v in poly.items():
        pol1 = moment_v_theta(k[4])  # ['k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
        pol2 = key_times_poly(v, k, pol1)
        poln.merge(pol2)
    return poln

def poly2num(poly, par):
    # kf = ['e^{kt}', 't', 'k^{-}',
    #       'mu', 'E[v]', 'theta', 'sigma', 'rho',
    #       'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s']
    k = par['k']
    mu, theta, sigma, rho = par['mu'], par['theta'], par['sigma'], par['rho']
    lmbd, mu_v, rhoJ, mu_s, sigma_s = par['lmbd'], par['mu_v'], par['rhoJ'], par['mu_s'], par['sigma_s']
    h = par['h']
    #
    mean_v = theta + lmbd * mu_v / k
    #
    f = 0
    for K, V in poly.items():
        val  = math.exp(K[0]*k*h) * (h ** K[1]) / (k ** K[2])
        val *= (mu ** K[3]) * (mean_v ** K[4]) * (theta ** K[5])
        val *= (sigma ** K[6]) * (rho ** K[7])
        val *= (lmbd ** K[8]) * (mu_v ** K[9]) * (rhoJ ** K[10])
        val *= (mu_s ** K[11]) * (sigma_s ** K[12])
        f += val * V
    return f

def unm_y(n, par):
    moment = moment_y(n)
    value = poly2num(moment, par)
    return value


if __name__ == '__main__':
    from pprint import pprint

    n = 4
    moment = moment_y(n)
    print(f"moment_y({n}) =")
    pprint(moment)
    print(f"{len(moment)} rows of {moment.keyfor}")



"""
Central Moments of :math:`IEZ_{s,t}`
"""
import math
from fractions import Fraction

from ajdmom.poly import Poly
from ajdmom.mdl_srjd.iez_mom import moment_IEZ, poly2num

def cmoment_IEZ(n):
    r"""Central moment of :math:`E[\overline{IEZ}_{s,t}^n]`

    :param int n: order of the central moment
    :return: poly with ``keyfor`` = ('e^{kt}', 'e^{ks}', 'k^{-}', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    if n < 0:
        raise ValueError(f"moment_IEZ({n}) is called!")
    kf = ['e^{kt}', 'e^{ks}', 'k^{-}', 'lmbd', 'mu_v']
    poly = Poly()
    poly.set_keyfor(kf)
    #
    p0 = Poly({(0, 0, 0, 0, 0): Fraction(0, 1)})
    p0.set_keyfor(kf)
    p1 = Poly({(0, 0, 0, 0, 0): Fraction(1, 1)})
    p1.set_keyfor(kf)
    if n == 0:
        return p1
    if n == 1:
        return p0
    mean = moment_IEZ(1)
    for i in range(n + 1):
        sign = (-1) ** i
        bino = math.comb(n, i)
        pol1 = p1 if i == 0 else mean ** i
        poln = (sign * bino) * (pol1 * moment_IEZ(n - i))
        poly.merge(poln)
    return poly


def cm(n, par):
    cmoment = cmoment_IEZ(n)
    value = poly2num(cmoment)
    return value


if __name__ == '__main__':
    from pprint import pprint

    print('\nExample usage of the module functions')
    n = 3
    print(f"cmoment_IEZ({n}) = ")
    pprint(cmoment_IEZ(n))
    print("which is a poly with keyfor = ('e^{kt}', 'e^{ks}', 'k^{-}', 'lmbd', 'mu_v')")



"""
Moments of the SVVJ model
"""
import math

from ajdmom import Poly
from ajdmom.mdl_svvj.cond_mom import moment_y as cond_moment_y
from ajdmom.mdl_svcj.mom import moment_v_theta

def key_times_poly(coef, k, poly):
    # k: ['e^{kt}', 't', 'k^{-}',
    #     'mu', 'v0-theta', 'theta', 'sigma', 'rho',
    #            substituted
    #     'lmbd', 'mu_v']
    #
    # poly: ['k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    #
    kf = ['e^{kt}', 't', 'k^{-}',
          'mu', 'E[v]', 'theta', 'sigma', 'rho',
          'lmbd', 'mu_v']
    poln = Poly()
    poln.set_keyfor(kf)
    for K, V in poly.items():
        key = (k[0],         # e^{kt}
               k[1],         # t
               k[2] + K[0],  # k^{-}
               k[3],         # mu
                      K[1],  # E[v]
               k[5],         # theta
               k[6] + K[2],  # sigma
               k[7],         # rho
               k[8] + K[3],  # lmbd
               k[9] + K[4]   # mu_v
               )
        poln.add_keyval(key, coef * V)
    return poln

def moment_y(n):
    """moment of :math:`y_t`

    :param int n: order of the moment
    :return: poly with ``keyfor`` = ('e^{kt}', 't', 'k^{-}',
      'mu', 'E[v]', 'theta', 'sigma', 'rho',
      'lmbd', 'mu_v')
    :rtype: Poly
    """
    poly = cond_moment_y(n)
    # kf = ['e^{kt}', 't', 'k^{-}',
    #       'mu', 'v0-theta', 'theta', 'sigma', 'rho',
    #       'lmbd', 'mu_v']
    kf = ['e^{kt}', 't', 'k^{-}',
          'mu', 'E[v]', 'theta', 'sigma', 'rho',
          'lmbd', 'mu_v']
    poln = Poly()
    poln.set_keyfor(kf)
    #
    for k, v in poly.items():
        pol1 = moment_v_theta(k[4])  # ['k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
        pol2 = key_times_poly(v, k, pol1)
        poln.merge(pol2)
    return poln

def expand_Ev(poly):
    # kf = ['e^{kt}', 't', 'k^{-}', 'mu', 'E[v]', 'theta', 'sigma', 'rho', 'lmbd', 'mu_v']
    kf = ['e^{kt}', 't', 'k^{-}', 'mu',           'theta', 'sigma', 'rho', 'lmbd', 'mu_v']
    poln = Poly()
    poln.set_keyfor(kf)
    for k, v in poly.items():
        for i in range(k[4] + 1):
            j = k[4] - i
            bino = math.comb(k[4], i)
            key = k[0:2] + (k[2]+j, k[3]) + k[5:8] + (k[8]+j, k[9]+j)
            val = bino * v
            poln.add_keyval(key, val)
    return poln

def poly2num(poly, par):
    # kf = ['e^{kt}', 't', 'k^{-}',
    #       'mu', 'E[v]', 'theta', 'sigma', 'rho',
    #       'lmbd', 'mu_v']
    k = par['k']
    mu, theta, sigma, rho = par['mu'], par['theta'], par['sigma'], par['rho']
    lmbd, mu_v = par['lmbd'], par['mu_v']
    h = par['h']
    #
    mean_v = theta + lmbd * mu_v / k
    #
    f = 0
    for K, V in poly.items():
        val  = math.exp(K[0]*k*h) * (h ** K[1]) / (k ** K[2])
        val *= (mu ** K[3]) * (mean_v ** K[4]) * (theta ** K[5])
        val *= (sigma ** K[6]) * (rho ** K[7])
        val *= (lmbd ** K[8]) * (mu_v ** K[9])
        f += val * V
    return f

def unm_y(n, par):
    moment = moment_y(n)
    value = poly2num(moment, par)
    return value

if __name__ == '__main__':
    from pprint import pprint

    n = 2
    moment = moment_y(n)
    moment = expand_Ev(moment)
    print(f"moment_y({n}) =")
    pprint(moment)
    print(f"{len(moment)} rows of {moment.keyfor}")
